from __future__ import annotations

import ray
from loguru import logger
from vllm import RequestOutput, SamplingParams

from rlite.interface.base_engine import BaseEngine, abort_if_no_executor
from rlite.interface.inference.inference_executor import (
    INFERENCE_EXECUTOR_REGISTRY,
    BaseInferenceExecutor
)
from rlite.nn.inference_module import BaseInferenceModule
from rlite.utils.need_parallel import NeedParallel


class RliteInferenceEngine(BaseEngine):
    def __init__(
        self,
        module_or_path: str | BaseInferenceModule,
        executor: BaseInferenceExecutor | str = "vllm",
        **kwargs
    ):
        super().__init__(**kwargs)

        if isinstance(module_or_path, BaseInferenceModule):
            self._module_path = module_or_path.get_checkpoint_path()
        else:
            self._module_path = module_or_path

        if isinstance(executor, str):
            self._executor_cls = INFERENCE_EXECUTOR_REGISTRY[executor]
        else:
            self._executor_cls = executor

    @property
    def executor_cls(self) -> type[BaseInferenceExecutor]:
        return self._executor_cls

    @abort_if_no_executor
    def generate(
        self,
        prompts: list[str],
        sampling_params: dict | SamplingParams | None = None,
        *,
        use_tqdm: bool = True,
        tqdm_desc: str = "Processed prompts"
    ) -> list[RequestOutput]:
        # RliteInferenceEngine.generate() will spread the prompts evenly to all
        # the workers. It is the caller's responsibility to handle the load
        # balance issue.

        if isinstance(sampling_params, dict):
            sampling_params = SamplingParams(**sampling_params)

        if isinstance(prompts, NeedParallel):
            prompts = prompts.unwrap()
            logger.info(
                "NOTE: rlite by default uses data parallel during "
                "inference, and the sequence parallel is handled by the "
                "backend inference engine."
            )

        # Assign prompts to each dp group.
        # Example 1:
        #   len(prompts) = 10, num_dp_groups = 3
        #   num_prompts_for_each_dp = [4, 3, 3]
        # Example 2:
        #   len(prompts) = 2, num_dp_groups = 4
        #   num_prompts_for_each_dp = [1, 1, 0, 0]
        num_dp_groups = len(self.executors)
        num_prompts_for_each_dp = [len(prompts) // num_dp_groups] * num_dp_groups
        for i in range(len(prompts) % num_dp_groups):
            num_prompts_for_each_dp[i] += 1
        dp_group_prompt_indeces = [0] + num_prompts_for_each_dp
        for i in range(1, len(dp_group_prompt_indeces)):
            dp_group_prompt_indeces[i] += dp_group_prompt_indeces[i - 1]

        prompts_for_each_dp = [
            prompts[dp_group_prompt_indeces[i]:dp_group_prompt_indeces[i + 1]]
            for i in range(num_dp_groups)
        ]

        # Generate responses from each dp group
        futures = [
            executor.generate(
                prompts=prompts_for_each_dp[i],
                sampling_params=sampling_params,
                use_tqdm=use_tqdm,
                tqdm_desc=tqdm_desc
            )
            for i, executor in enumerate(self.executors)
        ]
        responses: list[list[RequestOutput]] = [ray.get(future) for future in futures]

        # Deduplicate responses, for each executor, only the response from
        # the first worker is used, as all workers are running the same
        # prompt.
        deduped_responses = sum([x[0] for x in responses], [])
        return deduped_responses

    def pre_build_executors_hook(self, **kwargs) -> dict:
        return {"model_path": self._module_path}



import torch
from accelerate import init_empty_weights
from peft import LoraConfig, TaskType, get_peft_model
from torch.distributed.tensor import DTensor
from transformers import Qwen2Config, Qwen2ForCausalLM

import rlite
import rlite.nn
from rlite import RliteInferenceEngine, RliteTrainEngine


class MyModel(rlite.nn.HuggingFaceFsdp2TrainModule, Qwen2ForCausalLM):
    def on_weights_materialized(self):
        self._init_lora_B_as_zeros()

    def preprocess_input_named_parameters(
        self,
        name: str,
        param: torch.Tensor,
    ) -> tuple[str, torch.Tensor]:
        if not hasattr(self, "weight_mapping"):
            raise ValueError("weight_mapping is not set")
        if name in self.weight_mapping:
            name = self.weight_mapping[name]
        return name, param

    def postprocess_output_named_parameters(
        self,
        name: str,
        param: DTensor,
        full_state_dict: dict[str, DTensor]
    ) -> tuple[str, torch.Tensor]:
        if "base_layer.weight" in name:
            lora_a = full_state_dict[name.replace("base_layer.", "lora_A.default.")]
            lora_b = full_state_dict[name.replace("base_layer.", "lora_B.default.")]
            param = param.full_tensor() + lora_b.full_tensor() @ lora_a.full_tensor()
        else:
            param = param.full_tensor()
        name = name.replace("base_model.model.", "").replace("base_layer.", "")
        return name, param

    def _init_lora_B_as_zeros(self):
        with torch.no_grad():
            for name, param in self.named_parameters():
                if "lora_B" in name:
                    param.zero_()


def apply_lora(model: Qwen2ForCausalLM) -> torch.nn.Module:
    peft_confg = LoraConfig(
        r=32,
        lora_alpha=128,
        target_modules="all-linear",
        lora_dropout=0,
        bias="none",
        task_type=TaskType.CAUSAL_LM,
    )
    weight_mapping = {}
    orig_state_dict = model.state_dict()
    model = get_peft_model(model, peft_confg)
    new_state_dict = model.state_dict()
    for key in new_state_dict.keys():
        orig_key = key.replace("base_model.model.", "").replace("base_layer.", "")
        if orig_key in orig_state_dict:
            weight_mapping[orig_key] = key
    model.weight_mapping = weight_mapping

    return model


if __name__ == "__main__":
    rlite.init()

    model_name_or_path = "Qwen/Qwen2.5-7B-Instruct"

    vllm_engine = RliteInferenceEngine(model_name_or_path, executor="vllm")
    vllm_engine.build(tensor_parallel_size=4)

    vllm_engine.meta()  # Drop all the weights

    with init_empty_weights():
        config = Qwen2Config.from_pretrained(model_name_or_path)
        model = apply_lora(MyModel(config))

    fsdp2_engine = RliteTrainEngine(model, executor="fsdp2")
    fsdp2_engine.build(tensor_parallel_size=8, colocate_with=vllm_engine)

    print("Syncing weights from VLLM to FSDP2", flush=True)
    vllm_engine.cuda("weights")
    fsdp2_engine.p2p_weight_sync(vllm_engine)
    fsdp2_engine.cpu()  # Offload the weights to CPU
    vllm_engine.cuda("kv_cache")

    prompts = [
        "你好，请介绍一下你自己。",
        "Hello, how are you?",
        "中国的首都是",
        "France is famous for its",
    ]
    print(vllm_engine.generate(prompts))



from importlib.metadata import PackageNotFoundError, version

from packaging import version as pkg_version


def get_version(pkg):
    try:
        return version(pkg)
    except PackageNotFoundError:
        return None


package_name = "vllm"
package_version = get_version(package_name)


supported_versions = ["0.8.3", "0.8.4", "0.8.5", "0.8.5.post1"]


if package_version not in supported_versions:
    raise ImportError(f"vllm 版本 {package_version} 不支持。支持的版本为: {supported_versions}。")


if package_version in ["0.8.3", "0.8.4"]:
    from .vllm_0_8_34 import LLM

elif package_version in ["0.8.5", "0.8.5.post1"]:
    from .vllm_0_8_5 import LLM


__all__ = [
    "LLM",
]



from __future__ import annotations


class Registry:
    def __init__(self, name: str):
        self._name = name
        self._registry = {}

    def register(self, names: list[str] | str | None = None):
        if names is None:
            names = []
        elif isinstance(names, str):
            names = [names]

        def decorator(cls):
            for name in names:
                self._registry[name] = cls
            return cls

        return decorator

    def get(self, name: str, default=None) -> type:
        return self._registry.get(name, default)

    def __getitem__(self, name):
        try:
            return self._registry[name]
        except KeyError:
            raise KeyError(f"{name} is not registered in Registry {self._name}")



import abc
from functools import partial

from torch.distributed.fsdp.wrap import transformer_auto_wrap_policy

from rlite.nn.train_module import BaseHuggingFaceTrainModule, BaseTrainModule


class FsdpTrainModule(BaseTrainModule, abc.ABC):
    @abc.abstractmethod
    def get_fsdp_wrap_policy(self):
        """Get the FSDP wrap policy.

        This method is called by the `FsdpRayTrainDataParallel` class to get
        the FSDP auto wrap policy.
        """
        raise NotImplementedError()


class HuggingFaceFsdpTrainModule(BaseHuggingFaceTrainModule, FsdpTrainModule):
    """Rlite FSDP Module for Hugging Face models."""

    def get_fsdp_wrap_policy(self):
        if len(self.non_split_modules) == 0:
            raise NotImplementedError()
        return partial(transformer_auto_wrap_policy, transformer_layer_cls=self.non_split_modules)



import torch
from vllm import RequestOutput, SamplingParams

from rlite.interface.inference.inference_worker import BaseWorker
from rlite.third_party.vllm import LLM
from rlite.utils.distributed import CUDAIPCHandle


class VllmWorker(BaseWorker):
    def __init__(
        self,
        *,
        model_path: str,
        seed: int,
        tensor_parallel_size: int = 1,
        pipeline_parallel_size: int = 1,
        dtype: str = "bfloat16",
        enforce_eager: bool = True,
        gpu_memory_utilization: float = 0.8,
        trust_remote_code: bool = True,
        enable_chunked_prefill: bool = True,
        enable_prefix_caching: bool = True,
        max_model_len: int = 32768,
        use_sleep_mode: bool = False,
        **vllm_kwargs
    ):
        # Discard non-vllm kwargs
        for key in ["dp_rank", "data_parallel_size"]:
            vllm_kwargs.pop(key, None)

        self.llm = LLM(
            # These parameters are fixed
            enable_sleep_mode=use_sleep_mode,
            disable_custom_all_reduce=True,
            distributed_executor_backend="external_launcher",
            # Customizable parameters
            model=model_path,
            seed=seed,
            tensor_parallel_size=tensor_parallel_size,
            pipeline_parallel_size=pipeline_parallel_size,
            dtype=dtype,
            enforce_eager=enforce_eager,
            gpu_memory_utilization=gpu_memory_utilization,
            trust_remote_code=trust_remote_code,
            enable_chunked_prefill=enable_chunked_prefill,
            enable_prefix_caching=enable_prefix_caching,
            max_model_len=max_model_len,
            **vllm_kwargs
        )
        # NOTE: the GPU memory of vLLM is controlled by cuMemAllocator,
        #       which bypasses the pytorch's memory allocator. So the device
        #       of the model weight will always be cuda, even if the memory is
        #       released. To address this issue, we manually record the current
        #       device.
        self._device = "cuda"

    def generate(
        self,
        prompts: str | list[str],
        sampling_params: SamplingParams | dict | None = None,
        use_tqdm: bool = True,
        tqdm_desc: str = "Processed prompts",
        return_none: bool = False,
    ) -> list[RequestOutput]:
        if isinstance(prompts, str):
            prompts = [prompts]

        if len(prompts) == 0:
            return []

        if sampling_params is None:
            sampling_params = self.llm.get_default_sampling_params()
        elif isinstance(sampling_params, dict):
            sampling_params = SamplingParams(**sampling_params)

        outputs = self.llm.generate(
            prompts=prompts,
            sampling_params=sampling_params,
            use_tqdm=use_tqdm,
            tqdm_desc=tqdm_desc,
        )

        if return_none:
            return
        return outputs

    def cpu(self):
        self.llm.sleep(level=1)
        self.release_memory_cache(gpu=True)
        self._device = "cpu"

    def meta(self):
        self.llm.sleep(level=2)
        self.release_memory_cache(gpu=True)
        self._device = "meta"

    def cuda(self, tags: str | list[str] | None = None):
        if tags is None:
            tags = ["weights", "kv_cache"]
        if isinstance(tags, str):
            tags = [tags]
        if isinstance(tags, list):
            tags = set(tags)

        self.llm.wake_up(tags=tags)
        self._device = "cuda"

    def get_device(self) -> str:
        return self._device

    def update_weight(self, key: str, value: torch.Tensor | CUDAIPCHandle | None = None):
        if key == "END":
            return

        model = self.llm.llm_engine.model_executor.driver_worker.worker.model_runner.model

        if isinstance(value, CUDAIPCHandle):
            value = value.rebuild()

        try:
            model.load_weights(weights=[(key, value)])
        except KeyError:
            # Ignore the unrecognized weights.
            pass

    def load_state_dict(self, state_dict: dict):
        model = self.llm.llm_engine.model_executor.driver_worker.worker.model_runner.model
        model.load_weights(weights=state_dict.items())

    def state_dict(self) -> dict:
        raise NotImplementedError("Currently VllmWorker does not support state_dict()")

    def sync_weights_to(self, worker):
        raise NotImplementedError("Currently VllmWorker does not support sync_weights_to()")



from rlite.interface.train.train_executor import (
    TRAIN_EXECUTOR_REGISTRY,
    BaseTrainExecutor
)
from rlite.interface.train.train_worker import BaseTrainWorker

__all__ = [
    "TRAIN_EXECUTOR_REGISTRY",
    "BaseTrainExecutor",
    "BaseTrainWorker"
]



from accelerate import init_empty_weights
from transformers import Qwen2Config, Qwen2ForCausalLM

import rlite
import rlite.nn
from rlite import RliteInferenceEngine, RliteTrainEngine


class MyModel(rlite.nn.HuggingFaceFsdp2TrainModule, Qwen2ForCausalLM):
    ...


if __name__ == "__main__":
    rlite.init()

    model_name_or_path = "Qwen/Qwen2.5-7B-Instruct"

    vllm_engine = RliteInferenceEngine(model_name_or_path, executor="vllm")
    vllm_engine.build(tensor_parallel_size=4)

    vllm_engine.meta()  # Drop all the weights

    with init_empty_weights():
        config = Qwen2Config.from_pretrained(model_name_or_path)
        model = MyModel(config)

    fsdp2_engine = RliteTrainEngine(model, executor="fsdp2")
    fsdp2_engine.build(tensor_parallel_size=8, colocate_with=vllm_engine)

    vllm_engine.cuda("weights")
    fsdp2_engine.p2p_weight_sync(vllm_engine)
    fsdp2_engine.cpu()  # Offload the weights to CPU
    vllm_engine.cuda("kv_cache")

    prompts = [
        "你好，请介绍一下你自己。",
        "Hello, how are you?",
        "中国的首都是",
        "France is famous for its",
    ]
    print(vllm_engine.generate(prompts))



from .resource_bundle import ResourceBundle
from .resource_consumer import ResourceConsumer
from .resource_manager import RESMAN, ResourceManager


def available_gpus() -> list[int]:
    return RESMAN.available_resource_bundles(resource_type="gpu")


def total_gpus() -> int:
    return len(RESMAN.resource_bundles)


__all__ = [
    "ResourceBundle",
    "ResourceConsumer",
    "ResourceManager",
    "available_gpus",
    "total_gpus",
]



from __future__ import annotations

import abc
import os
from functools import cached_property
from pathlib import Path

import torch

from rlite.utils.distributed import SerializableModule


class BaseTrainModule(SerializableModule, abc.ABC):
    def init_hook(self):
        """
        Hook that is called to transform the model before it is bind to the
        engine.
        """
        return self

    def materialization_source(self) -> dict[str, torch.Tensor] | dict[str, callable] | str:
        """Materialization source.

        Returns:
            source: The source of the weights. Can be a state dict (unsharded
                    weights on CPU), a dict of `param_init_fn` (iteratively
                    initialize the weights), or a path to the checkpoint files.
        """
        raise NotImplementedError()

    def on_weights_materialized(self):
        """Hook that is called after the model's weights are materialized.

        This method is designed for logic that happens after the model's
        weights are materialized on the GPU.
        """

    def preprocess_input_named_parameters(
        self,
        name: str,
        param: torch.Tensor
        # NOTE: should we support the full state dict in this function?
    ) -> tuple[str, torch.Tensor]:
        """Preprocess the input named parameters.

        This function is called whenever the model loads weight from outside.
        Examples include loading checkpoints from the file system or receiving
        weight from other workers.
        """
        return name, param  # By default, do nothing.

    def postprocess_output_named_parameters(
        self,
        name: str,
        param: torch.Tensor,
        full_state_dict: dict[str, torch.Tensor]
    ) -> tuple[str, torch.Tensor]:
        """Postprocess the output named parameters.

        This function is called whenever the model saves weight to outside.
        Examples include saving checkpoints to the file system or sending
        weights to other workers.
        """
        return name, param  # By default, do nothing.


class BaseHuggingFaceTrainModule(BaseTrainModule):
    def init_hook(self):
        """Infer the materialization source and download weights if needed."""

        def get_latest_directory(directory: Path) -> Path | None:
            if not directory.exists():
                return None
            directories = [d for d in directory.iterdir() if d.is_dir()]
            if not directories:
                return None
            return max(directories, key=lambda d: d.stat().st_mtime)

        if Path(self.config.name_or_path).exists():
            self._materialization_source = self.config.name_or_path
            return

        # Infer the default path for cache
        default_cache_root = Path.home() / ".cache/huggingface/hub"
        cache_root = Path(os.getenv("HF_HOME", os.getenv("HF_HUB_CACHE", default_cache_root)))
        cache_name = self.config.name_or_path.replace("/", "--")
        cache_path = get_latest_directory(cache_root / f"models--{cache_name}" / "snapshots")

        # Download if no safetensors found (TODO: support other formats)
        if len(list(cache_path.glob("*.safetensors"))) == 0:  # No safetensors found, download
            import huggingface_hub
            huggingface_hub.snapshot_download(self.config.name_or_path)
            cache_path = get_latest_directory(cache_root / f"models--{cache_name}" / "snapshots")
            if cache_path is None:
                raise ValueError(f"Failed to download the model from {self.config.name_or_path}")

        self._materialization_source = str(cache_path)

    def materialization_source(self) -> str:
        """Always return an existing path in the file system."""
        return self._materialization_source

    @cached_property
    def non_split_modules(self) -> set[type[torch.nn.Module]]:
        if hasattr(self, "_no_split_modules"):
            class_names = self._no_split_modules
            no_split_modules = set()
            for module in self.modules():
                if module.__class__.__name__ in class_names:
                    no_split_modules.add(module.__class__)
            return no_split_modules

        elif hasattr(self, "get_decoder"):
            return {self.get_decoder().layers[0].__class__}

        elif hasattr(self, "layers"):
            return {self.layers[0].__class__}

        return set()



from __future__ import annotations

from typing import TYPE_CHECKING

import ray
from ray.util.placement_group import PlacementGroup
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy

if TYPE_CHECKING:
    from rlite.resman import ResourceManager


class ResourceBundle:
    """A resource bundle that contains the resource request of a consumer."""

    def __init__(self, num_cpus: int, num_gpus: float):
        self.num_cpus = num_cpus
        self.num_gpus = num_gpus
        self.__total_num_cpus = num_cpus
        self.__total_num_gpus = num_gpus

        # Resource allocation info
        self.bundle_id_in_placement_group = None
        self.node_id = None
        self.device_id = None
        self.placement_group = None
        self.resource_manager = None

    @property
    def total_num_cpus(self) -> int:
        return self.__total_num_cpus

    @property
    def total_num_gpus(self) -> int:
        return self.__total_num_gpus

    def to_dict(self, include_device_info: bool = False) -> dict:
        if include_device_info and self.is_resource_allocated:
            ans = {
                "node_id": self.node_id,
                "device_id": self.device_id,
                "CPU": self.num_cpus,
                "GPU": self.num_gpus,
            }
        else:
            ans = {"CPU": self.num_cpus, "GPU": self.num_gpus}

        return {k: v for k, v in ans.items() if v is not None}

    def bind_resources(
        self,
        *,
        device_id: int,
        bundle_id: int,
        node_id: str,
        placement_group: PlacementGroup,
        resource_manager: ResourceManager,
    ):
        if not isinstance(node_id, str) or len(node_id) == 0:
            raise ValueError("Invalid node ID!", node_id)
        if not isinstance(device_id, int) or device_id < 0:
            raise ValueError("Invalid device ID!", device_id)
        if not isinstance(bundle_id, int) or bundle_id < 0:
            raise ValueError("Invalid bundle ID!", bundle_id)
        if not isinstance(placement_group, PlacementGroup):
            raise ValueError("Invalid placement group!", placement_group)

        self.device_id = device_id
        self.bundle_id_in_placement_group = bundle_id
        self.node_id = node_id
        self.placement_group = placement_group
        self.resource_manager = resource_manager

    def remote(self, actor: any, **options):
        if not self.is_resource_allocated:
            raise RuntimeError("This resource bundle's resources are not allocated!")
        options.update(
            {
                "scheduling_strategy": PlacementGroupSchedulingStrategy(
                    placement_group=self.placement_group,
                    placement_group_bundle_index=self.bundle_id_in_placement_group,
                ),
                "num_cpus": self.num_cpus,
                "num_gpus": self.num_gpus,
            }
        )
        return ray.remote(actor).options(**options)

    @property
    def is_resource_allocated(self) -> bool:
        return self.placement_group is not None

    def __add__(self, other: "ResourceBundle") -> "ResourceBundle":
        if self.is_resource_allocated or other.is_resource_allocated:
            if self.node_id != other.node_id or self.device_id != other.device_id:
                raise RuntimeError("Cannot add resource bundles from different nodes or devices!")

        new_bundle = ResourceBundle(
            num_cpus=self.num_cpus + other.num_cpus,
            num_gpus=self.num_gpus + other.num_gpus,
        )
        if self.is_resource_allocated:
            new_bundle.bind_resources(
                bundle_id=self.bundle_id_in_placement_group,
                node_id=self.node_id,
                device_id=self.device_id,
                placement_group=self.placement_group,
                resource_manager=self.resource_manager,
            )
        return new_bundle

    def __sub__(self, other: "ResourceBundle") -> "ResourceBundle":
        if self.is_resource_allocated and other.is_resource_allocated:
            if self.node_id != other.node_id or self.device_id != other.device_id:
                raise RuntimeError(
                    "Cannot subtract resource bundles from different nodes or devices!"
                )

        new_bundle = ResourceBundle(
            num_cpus=self.num_cpus - other.num_cpus,
            num_gpus=self.num_gpus - other.num_gpus,
        )
        if self.is_resource_allocated:
            new_bundle.bind_resources(
                bundle_id=self.bundle_id_in_placement_group,
                node_id=self.node_id,
                device_id=self.device_id,
                placement_group=self.placement_group,
                resource_manager=self.resource_manager,
            )
        return new_bundle

    def __iadd__(self, other: "ResourceBundle") -> "ResourceBundle":
        if self.is_resource_allocated or other.is_resource_allocated:
            if self.node_id != other.node_id or self.device_id != other.device_id:
                raise RuntimeError("Cannot add resource bundles from different nodes or devices!")

        self.num_cpus += other.num_cpus
        self.num_gpus += other.num_gpus
        return self

    def __isub__(self, other: "ResourceBundle") -> "ResourceBundle":
        if self.is_resource_allocated and other.is_resource_allocated:
            if self.node_id != other.node_id or self.device_id != other.device_id:
                raise RuntimeError(
                    "Cannot subtract resource bundles from different nodes or devices!"
                )

        self.num_cpus -= other.num_cpus
        self.num_gpus -= other.num_gpus
        return self

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ResourceBundle):
            raise NotImplementedError()
        return self.num_cpus == other.num_cpus and self.num_gpus == other.num_gpus

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, ResourceBundle):
            raise NotImplementedError()
        return self.num_cpus > other.num_cpus and self.num_gpus > other.num_gpus

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, ResourceBundle):
            raise NotImplementedError()
        return self.num_cpus >= other.num_cpus and self.num_gpus >= other.num_gpus

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ResourceBundle):
            raise NotImplementedError()
        return not self >= other

    def __le__(self, other: object) -> bool:
        if not isinstance(other, ResourceBundle):
            raise NotImplementedError()
        return not self > other

    def __repr__(self) -> str:
        fields = ["node_id", "device_id", "num_cpus", "num_gpus"]
        formatted_fields = ", ".join(f"{field}={getattr(self, field)}" for field in fields)
        return f"ResourceBundle({formatted_fields})"



from __future__ import annotations

from collections import defaultdict

import torch
import torch.distributed as dist
import torch.nn as nn
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp import StateDictType
from torch.distributed.fsdp._runtime_utils import _lazy_init


def ensure_weights_retied(param_init_fn, model: torch.nn.Module, device: torch.cuda.device):
    """Handles `transformers` models with tied weights."""

    _tied_names = getattr(model, "_tied_weights_keys", None)
    if not _tied_names:
        # if no tied names just passthrough
        return param_init_fn

    # get map of parameter instances to params.
    # - needed for replacement later
    _tied_params = {}
    for name in _tied_names:
        name = name.split(".")
        name, param_name = ".".join(name[:-1]), name[-1]
        mod = model.get_submodule(name)
        param = getattr(mod, param_name)

        _tied_params[id(param)] = None  # placeholder for the param first

    # build param_init_fn for the case with tied params
    def param_init_fn_tied_param(module: torch.nn.Module):
        # track which params to tie
        # - usually only 1, but for completeness consider > 1
        params_to_tie = defaultdict(list)
        for n, param in module.named_parameters(recurse=False):
            if id(param) in _tied_params:
                params_to_tie[id(param)].append(n)

        # call the param init fn, which potentially re-allocates the
        # parameters
        module = param_init_fn(module)

        # search the parameters again and tie them up again
        for id_key, _param_names in params_to_tie.items():
            for param_name in _param_names:
                param = _tied_params[id_key]
                if param is None:
                    # everything will be tied to the first time the
                    # param is observed
                    _tied_params[id_key] = getattr(module, param_name).clone()
                else:
                    setattr(module, param_name, nn.Parameter(param))  # tie

        return module

    return param_init_fn_tied_param


@torch.no_grad()
def offload_model(model: torch.nn.Module):
    assert isinstance(model, FSDP)
    _lazy_init(model, model)
    assert model._is_root, "Only support root model offloading to CPU"
    for handle in model._all_handles:
        if handle._offload_params:
            continue
        flat_param = handle.flat_param
        assert (
            flat_param.data.data_ptr() == flat_param._local_shard.data_ptr()
            and id(flat_param.data) != id(flat_param._local_shard)
            and flat_param.data.size() == flat_param._local_shard.size()
        )
        handle.flat_param_to(torch.device("cpu"), non_blocking=True)
        flat_param._local_shard = flat_param.data
    torch.cuda.empty_cache()


@torch.no_grad()
def reload_model(model: FSDP):
    assert isinstance(model, FSDP)
    _lazy_init(model, model)
    assert model._is_root, "Only support root model loading to GPU"
    device_id = torch.cuda.current_device()
    for handle in model._all_handles:
        if handle._offload_params:
            continue
        flat_param = handle.flat_param
        handle.flat_param_to(torch.device(f"cuda:{device_id}"), non_blocking=True)
        flat_param._local_shard = flat_param.data


@torch.no_grad()
def offload_optimizer(optimizer: torch.optim.Optimizer):
    if not optimizer.state:
        return
    for param_group in optimizer.param_groups:
        for param in param_group["params"]:
            state = optimizer.state[param]
            for key, value in state.items():
                if isinstance(value, torch.Tensor):
                    state[key] = value.to("cpu", non_blocking=True)


def reload_optimizer(optimizer: torch.optim.Optimizer):
    if not optimizer.state:
        return
    for param_group in optimizer.param_groups:
        for param in param_group["params"]:
            state = optimizer.state[param]
            for key, value in state.items():
                if isinstance(value, torch.Tensor):
                    state[key] = value.to("cuda", non_blocking=True)


def iter_fsdp1_state_dict(model: FSDP | None = None, sharded_state_dict: dict | None = None):
    """Iterate over the state dict of an FSDP model.

    This function will gather the weights key by key, avoiding materializing
    the entire state dict at once and thus is memory efficient.
    """
    # NOTE (zhanghan): We used ShardedTensor.gather and dist.broadcast, where
    #                  we assume the model's sharding process group is built
    #                  with dist.init_process_group and its replicate group
    #                  is not. This is the default implementation in Rlite
    #                  parallel, since Rlite_init_process_group will lead to
    #                  hanging when initializing the FSDP instance.
    if model is None and sharded_state_dict is None:
        raise ValueError("Either `model` or `sharded_state_dict` must be provided.")

    if sharded_state_dict is None:
        with FSDP.state_dict_type(model, StateDictType.SHARDED_STATE_DICT):
            sharded_state_dict = model.state_dict()

    for k, v in sharded_state_dict.items():
        gathered_value = torch.empty(
            v.shape,
            device=torch.device("cuda"),
            dtype=v.dtype,
        )
        v.gather(out=gathered_value if dist.get_rank() == 0 else None)
        dist.broadcast(gathered_value, src=0)
        yield k, gathered_value


def param_init_preserve_buffers(module: torch.nn.Module):
    """Initialize the parameters of the module, but preserve the buffers."""
    buffers = {}
    for name, param in module.named_buffers():
        buffers[name] = param
    module = module.to_empty(device=torch.device("cuda"), recurse=False)
    for name, param in module.named_buffers():
        param.data.copy_(buffers[name].data)
    return module



from __future__ import annotations

import abc

import ray

from rlite.interface.base_worker import BaseWorker


class BaseExecutor(abc.ABC):
    # ----- Interfaces ----- #

    @property
    @abc.abstractmethod
    def workers(self) -> list[BaseWorker]:
        raise NotImplementedError()

    @abc.abstractmethod
    def build_workers(self) -> list[ray.ObjectRef]:
        raise NotImplementedError()

    # ----- Shared methods ----- #

    def load_state_dict(self, state_dict: dict) -> list[ray.ObjectRef]:
        futures = [worker.load_state_dict.remote(state_dict) for worker in self.workers]
        return futures

    def state_dict(self) -> list[ray.ObjectRef]:
        futures = [worker.state_dict.remote() for worker in self.workers]
        return futures

    def cpu(self, *args, **kwargs) -> list[ray.ObjectRef]:
        futures = [worker.cpu.remote(*args, **kwargs) for worker in self.workers]
        return futures

    def meta(self, *args, **kwargs) -> list[ray.ObjectRef]:
        futures = [worker.meta.remote(*args, **kwargs) for worker in self.workers]
        return futures

    def cuda(self, *args, **kwargs) -> list[ray.ObjectRef]:
        futures = [worker.cuda.remote(*args, **kwargs) for worker in self.workers]
        return futures

    def to(self, device: str, *args, **kwargs) -> list[ray.ObjectRef]:
        """Move the workers to the given device."""
        assert device in ["cpu", "cuda", "meta"]
        if device == "cpu":
            return self.cpu(*args, **kwargs)
        elif device == "cuda":
            return self.cuda(*args, **kwargs)
        elif device == "meta":
            return self.meta(*args, **kwargs)

    def release_memory_cache(self, cpu: bool = False, gpu: bool = True) -> list[ray.ObjectRef]:
        """Release the memory cache."""
        futures = [worker.release_memory_cache.remote(cpu, gpu) for worker in self.workers]
        return futures

    def get_device(self) -> list[ray.ObjectRef]:
        futures = [worker.get_device.remote() for worker in self.workers]
        return futures



import abc

from vllm import RequestOutput, SamplingParams

from rlite.interface import BaseWorker


class BaseInferenceWorker(BaseWorker, abc.ABC):
    @abc.abstractmethod
    def generate(
        self,
        prompts: str | list[str],
        sampling_params: SamplingParams | dict | None = None,
        use_tqdm: bool = True,
        tqdm_desc: str = "Processed prompts"
    ) -> list[RequestOutput]:
        raise NotImplementedError()



from __future__ import annotations

import ray

from rlite.interface.base_engine import BaseEngine, abort_if_no_executor
from rlite.interface.train import TRAIN_EXECUTOR_REGISTRY, BaseTrainExecutor
from rlite.nn import BaseTrainModule
from rlite.train.utils import prepare_parallel_args, prepare_parallel_kwargs


class RliteTrainEngine(BaseEngine):
    def __init__(
        self,
        module: BaseTrainModule,
        *,
        executor: BaseTrainExecutor | str = "fsdp2",
        **kwargs
    ):
        super().__init__(**kwargs)

        # assert isinstance(module, BaseTrainModule)
        assert next(module.parameters()).is_meta, "Use meta model to initialize!"
        self._module = module

        # Call the init_hook of the module
        self._module.init_hook()

        if isinstance(executor, str):
            self._executor_cls = TRAIN_EXECUTOR_REGISTRY[executor]
        else:
            self._executor_cls = executor

    @property
    def executor_cls(self) -> type[BaseTrainExecutor]:
        return self._executor_cls

    @abort_if_no_executor
    def state_dict(self) -> dict:
        return ray.get(self.executors[0].state_dict())

    @abort_if_no_executor
    def load_state_dict(self, state_dict: dict):
        # NOTE: This function is expected to be slow. Use CUDAIPC + colocate
        #       to speed up the transfer.
        futures = sum([
            executor.load_state_dict(state_dict)
            for executor in self.executors
        ], [])
        return ray.get(futures)

    @abort_if_no_executor
    def train(self):
        """Switch to training mode."""
        futures = sum([executor.train() for executor in self.executors], [])
        return ray.get(futures)

    @abort_if_no_executor
    def eval(self):
        """Switch to evaluation mode."""
        futures = sum([executor.eval() for executor in self.executors], [])
        return ray.get(futures)

    @property
    @abort_if_no_executor
    def training(self) -> bool:
        is_training = ray.get(sum([executor.is_training() for executor in self.executors]))
        if all(is_training):
            return True
        elif not any(is_training):
            return False
        else:
            raise RuntimeError("Mixed training and evaluation executors!")

    def __getattr__(self, attr_name: str) -> any:
        try:
            return object.__getattribute__(self, attr_name)
        except AttributeError:

            # Abort if no executor is available.
            if self.executors is None or len(self.executors) == 0:
                raise AttributeError(f"No attribute {attr_name} found in the engine.")

            # Magic part.
            # This function allows dynamically calling methods defined in the
            # module on all executors.
            def run_executors(*args, **kwargs):
                # Note that we don't unwrap the args and kwargs here, as the
                # executor can also employ data parallel, e.g. FSDP.
                args_list = prepare_parallel_args(
                    args,
                    unwrap=False,
                    dp_size=len(self.executors)
                )
                kwargs_list = prepare_parallel_kwargs(
                    kwargs,
                    unwrap=False,
                    dp_size=len(self.executors)
                )
                futures = [
                    executor.run_workers(attr_name, *args_list[dp_rank], **kwargs_list[dp_rank])
                    for dp_rank, executor in enumerate(self.executors)
                ]
                return [ray.get(future) for future in futures]

            return run_executors

    def pre_build_executors_hook(
        self,
        *,
        tensor_parallel_size: int,
        pipeline_parallel_size: int,
        data_parallel_size: int,
        **kwargs,
    ) -> dict:
        from rlite.train.fsdp2.executor import Fsdp2TrainExecutor
        from rlite.train.fsdp.executor import FsdpTrainExecutor

        if self.executor_cls is FsdpTrainExecutor and data_parallel_size > 1:
            # Prepare replicate group master address and ports.
            num_gpus_per_dp_group = tensor_parallel_size * pipeline_parallel_size
            replicate_master_addrs = [
                self._resource_manager.get_ip_addr(self._resource_bundles[i])
                for i in range(num_gpus_per_dp_group)
            ]
            replicate_ports = [
                self._resource_manager.get_free_port(self._resource_bundles[i])
                for i in range(num_gpus_per_dp_group)
            ]
            return {
                "module": self._module,
                "replicate_master_addrs": replicate_master_addrs,
                "replicate_ports": replicate_ports,
            }

        elif self.executor_cls is Fsdp2TrainExecutor:
            master_addr = self._resource_manager.get_ip_addr(self._resource_bundles[0])
            master_port = self._resource_manager.get_free_port(self._resource_bundles[0])
            mesh_shape = (data_parallel_size, tensor_parallel_size * pipeline_parallel_size)
            return {
                "module": self._module,
                "master_addr": master_addr,
                "master_port": master_port,
                "mesh_shape": mesh_shape,
            }

        return {"module": self._module}

    def save(self, checkpoint_path: str, *args, **kwargs):
        ray.get(self.executors[0].save(checkpoint_path, *args, **kwargs))

    def load(self, checkpoint_path: str, *args, **kwargs):
        futures = sum([
            executor.load(checkpoint_path, *args, **kwargs)
            for executor in self.executors
        ], [])
        return ray.get(futures)



from rlite.nn.fsdp import FsdpTrainModule, HuggingFaceFsdpTrainModule
from rlite.nn.fsdp2 import Fsdp2TrainModule, HuggingFaceFsdp2TrainModule
from rlite.nn.inference_module import BaseInferenceModule
from rlite.nn.train_module import BaseHuggingFaceTrainModule, BaseTrainModule

__all__ = [
    "BaseTrainModule",
    "BaseHuggingFaceTrainModule",
    "BaseInferenceModule",
    "FsdpTrainModule",
    "HuggingFaceFsdpTrainModule",
    "Fsdp2TrainModule",
    "HuggingFaceFsdp2TrainModule",
]



from typing import Optional, Sequence, Union, cast

from tqdm import tqdm
from vllm import LLM as BaseLLM
from vllm.executor.uniproc_executor import ExecutorWithExternalLauncher
from vllm.inputs import PromptType
from vllm.model_executor.guided_decoding.guided_fields import (
    GuidedDecodingRequest
)
from vllm.outputs import RequestOutput


class LLM(BaseLLM):
    def generate(
        self,
        prompts=None,
        sampling_params=None,
        prompt_token_ids=None,
        use_tqdm=True,
        lora_request=None,
        prompt_adapter_request=None,
        guided_options_request=None,
        priority=None,
        tqdm_desc="Processed prompts",
    ):
        runner_type = self.llm_engine.model_config.runner_type
        if runner_type not in ["generate", "transcription"]:
            messages = [
                "LLM.generate() is only supported for (conditional) generation "
                "models (XForCausalLM, XForConditionalGeneration).",
            ]

            supported_runner_types = self.llm_engine.model_config \
                .supported_runner_types
            if "generate" in supported_runner_types:
                messages.append(
                    "Your model supports the 'generate' runner, but is "
                    f"currently initialized for the '{runner_type}' runner. "
                    "Please initialize vLLM using `--task generate`.")

            raise ValueError(" ".join(messages))

        if prompt_token_ids is not None:
            parsed_prompts = self._convert_v1_inputs(
                prompts=cast(Optional[Union[str, list[str]]], prompts),
                prompt_token_ids=prompt_token_ids,
            )
        else:
            parsed_prompts = cast(Union[PromptType, Sequence[PromptType]],
                                  prompts)

        if isinstance(guided_options_request, dict):
            if len(guided_options_request) > 1:
                raise ValueError(
                    "You can only use one guided decoding but multiple is "
                    f"specified: {guided_options_request}")
            guided_options_request = GuidedDecodingRequest(
                **guided_options_request)

        if sampling_params is None:
            # Use default sampling params.
            sampling_params = self.get_default_sampling_params()

        self._validate_and_add_requests(
            prompts=parsed_prompts,
            params=sampling_params,
            lora_request=lora_request,
            prompt_adapter_request=prompt_adapter_request,
            guided_options=guided_options_request,
            priority=priority)

        outputs = self._run_engine(use_tqdm=use_tqdm, tqdm_desc=tqdm_desc)
        return self.engine_class.validate_outputs(outputs, RequestOutput)

    def _run_engine(self, *, use_tqdm: bool, tqdm_desc: str):
        # Initialize tqdm.
        if use_tqdm:
            num_requests = self.llm_engine.get_num_unfinished_requests()
            pbar = tqdm(
                total=num_requests,
                desc=tqdm_desc,
                dynamic_ncols=True,
                postfix=(f"est. speed input: {0: .2f} toks/s, "
                         f"output: {0: .2f} toks/s"),
            )

        # Run the engine.
        outputs = []
        total_in_toks = 0
        total_out_toks = 0
        while self.llm_engine.has_unfinished_requests():
            step_outputs = self.llm_engine.step()
            for output in step_outputs:
                if output.finished:
                    outputs.append(output)
                    if use_tqdm:
                        if isinstance(output, RequestOutput):
                            # Calculate tokens only for RequestOutput
                            n = len(output.outputs)
                            assert output.prompt_token_ids is not None
                            total_in_toks += len(output.prompt_token_ids) * n
                            in_spd = total_in_toks / pbar.format_dict["elapsed"]
                            total_out_toks += sum(
                                len(stp.token_ids) for stp in output.outputs)
                            out_spd = (total_out_toks /
                                       pbar.format_dict["elapsed"])
                            pbar.postfix = (
                                f"est. speed input: {in_spd: .2f} toks/s, "
                                f"output: {out_spd: .2f} toks/s")
                            pbar.update(n)
                        else:
                            pbar.update(1)

        if use_tqdm:
            pbar.close()
        # Sort the outputs by request ID.
        # This is necessary because some requests may be finished earlier than
        # its previous requests.
        return sorted(outputs, key=lambda x: int(x.request_id))

    def sleep(self, level: int = 1):
        assert isinstance(self.llm_engine.model_executor, ExecutorWithExternalLauncher), \
            "In Rlite parallel, we only support ExecutorWithExternalLauncher!"

        self.reset_prefix_cache()

        if level == 2:
            model = self.llm_engine.model_executor.driver_worker.worker.model_runner.model
            self.saved_buffers_for_sleep = {
                name: buffer.cpu().clone()
                for name, buffer in model.named_buffers()
            }

        self.llm_engine.sleep(level=level)

    def wake_up(self, *args, **kwargs):
        self.llm_engine.wake_up(*args, **kwargs)

        if hasattr(self, "saved_buffers_for_sleep"):
            model = self.llm_engine.model_executor.driver_worker.worker.model_runner.model
            for name, buffer in model.named_buffers():
                buffer.data.copy_(self.saved_buffers_for_sleep[name])
            del self.saved_buffers_for_sleep



from rlite.interface.inference.inference_executor import (
    INFERENCE_EXECUTOR_REGISTRY,
    BaseInferenceExecutor
)
from rlite.interface.inference.inference_worker import BaseWorker
from rlite.nn.inference_module import BaseInferenceModule

__all__ = [
    "INFERENCE_EXECUTOR_REGISTRY",
    "BaseInferenceExecutor",
    "BaseWorker",
    "BaseInferenceModule",
]



from torch.utils.tensorboard import SummaryWriter


class DummySummaryWriter(SummaryWriter):
    def __getattr__(self, name: str):
        if name.startswith("add_"):
            return self.dummy_call
        return super().__getattr__(name)

    def dummy_call(self, *args, **kwargs):
        pass



import rlite.third_party.transformers.monkey_patch
from rlite.inference import RliteInferenceEngine
from rlite.resman import ResourceManager
from rlite.train import RliteTrainEngine
from rlite.utils import NeedParallel


def init(
    network_interface: str = "eth0",
    dedup_logs: bool = False,
    debug_mode: str = "debugpy",
    debug_post_mortem: bool = False,
):
    import os
    import socket

    import ray

    assert not ray.is_initialized(), "Let rlite initialize the ray!"

    assert debug_mode in ["debugpy", "pdb"], f"Invalid debug mode: {debug_mode}"

    ray_env = {"env_vars": {"RAY_DEDUP_LOGS": "0"}}

    if debug_mode == "pdb":
        os.environ["RAY_DEBUG"] = "legacy"
        ray_env["env_vars"]["RAY_DEBUG"] = "legacy"
        ray_env["env_vars"]["REMOTE_PDB_HOST"] = socket.gethostbyname(socket.gethostname())

    if debug_post_mortem:
        os.environ["RAY_DEBUG_POST_MORTEM"] = "1"
        ray_env["env_vars"]["RAY_DEBUG_POST_MORTEM"] = "1"

    if dedup_logs:
        ray_env["env_vars"]["RAY_DEDUP_LOGS"] = "1"

    ray.init(runtime_env=ray_env)

    import rlite.resman

    rlite.resman.RESMAN = ResourceManager(network_interface=network_interface)


__all__ = [
    "NeedParallel",
    "RliteTrainEngine",
    "RliteInferenceEngine",
    "ResourceManager"
]



from __future__ import annotations

from functools import cached_property, wraps
from typing import Generator

import ray
import torch
import torch.distributed as dist
import torch.nn as nn
from torch.distributed.device_mesh import DeviceMesh

from rlite.interface import BaseWorker
from rlite.interface.train import BaseTrainWorker
from rlite.nn import Fsdp2TrainModule
from rlite.train.utils.fsdp import offload_optimizer, reload_optimizer
from rlite.utils.checkpoint import load_safetensors_generator, save_safetensors
from rlite.utils.distributed import CUDAIPCHandle


def abort_if_not_cuda(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.get_device() != "cuda":
            raise RuntimeError(
                f"Currently the model is on {self.get_device()}, you need to call "
                "`cuda()` first to save the checkpoint."
            )
        return func(self, *args, **kwargs)
    return wrapper


def capture_forward_error(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RuntimeError as e:
            if "got mixed torch.Tensor and DTensor" in str(e):
                raise RuntimeError(
                    "FSDP2 does not support calling `forward` directly. That is, "
                    "calling `self.forward(...)` is expected to fail. Please call "
                    "`self(...)` instead."
                )
            raise e
    return wrapper


class Fsdp2Worker(BaseTrainWorker):
    def __init__(
        self,
        module: Fsdp2TrainModule,
        mesh_shape: tuple[int, ...],
        rank_in_shard_group: int,
        shard_group_world_size: int,
        **kwargs,
    ):
        self.module = module.float()
        self.rank_in_shard_group = rank_in_shard_group
        self.shard_group_world_size = shard_group_world_size
        self._init_model(self.module, mesh_shape)

    def __getattr__(self, name):
        if name == "model":
            return object.__getattribute__(self, "model")
        try:
            return super().__getattr__(name)
        except AttributeError:
            # MAGIC: Hijack the getattr to support any customized method.
            model = object.__getattribute__(self, "model")
            try:
                return nn.Module.__getattr__(model, name)
            except AttributeError:
                if hasattr(self.module, name):
                    attr = getattr(self.module, name)
                    if callable(attr):
                        return capture_forward_error(attr.__get__(model))
                raise

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    @property
    def training(self) -> bool:
        return self.model.training

    def cpu(self):
        self.model.cpu()
        if hasattr(self, "optimizer"):
            offload_optimizer(self.model.optimizer)
        self.release_memory_cache()

    def cuda(self):
        self.model.cuda()
        if hasattr(self, "optimizer"):
            reload_optimizer(self.model.optimizer)

    def meta(self):
        raise NotImplementedError("Currently not implemented, coming soon.")

    def get_device(self) -> str:
        return next(self.model.parameters()).device.type

    @abort_if_not_cuda
    def sync_weights_to(self, target_worker: ray.ActorHandle[BaseWorker]):
        for key, param in self._named_parameters.items():
            postprocess_fn = self.module.postprocess_output_named_parameters.__get__(self.model)
            key, param = postprocess_fn(key, param, self._named_parameters)
            ipc_handle = CUDAIPCHandle.from_tensor(param)
            future = target_worker.update_weight.remote(key, ipc_handle)
            ray.get(future)

        ray.get(target_worker.update_weight.remote("END"))

    @abort_if_not_cuda
    def update_weight(self, key: str, weight: torch.Tensor | CUDAIPCHandle | None = None):
        if key == "END":
            return
        if isinstance(weight, CUDAIPCHandle):
            weight = weight.rebuild()
        preprocess_fn = self.module.preprocess_input_named_parameters.__get__(self.model)
        key, weight = preprocess_fn(key, weight)
        self._named_parameters[key]._local_tensor.data.copy_(weight)

    @abort_if_not_cuda
    def state_dict(self) -> dict:
        local_state_dict = self.model.state_dict()
        postprocess_fn = self.module.postprocess_output_named_parameters.__get__(self.model)
        full_state_dict = dict([
            postprocess_fn(key, value, local_state_dict)
            for key, value in local_state_dict.items()
        ])
        if dist.get_rank() == 0:
            return full_state_dict

    @abort_if_not_cuda
    def load_state_dict(self, state_dict: dict | Generator):
        if isinstance(state_dict, dict):
            state_dict = state_dict.items()

        for key, value in state_dict:

            # Apply the preprocess function
            preprocess_fn = self.module.preprocess_input_named_parameters.__get__(self.model)
            key, value = preprocess_fn(key, value)

            # Chunk and load the weights
            chunked_value = torch.chunk(
                value, self.shard_group_world_size, dim=0
            )[self.rank_in_shard_group]
            self._named_parameters[key]._local_tensor.data.copy_(chunked_value)

    @abort_if_not_cuda
    def save(self, checkpoint_path: str):
        def _param_iterator():
            post_process_fn = self.module.postprocess_output_named_parameters.__get__(self.model)
            for name, param in self._named_parameters.items():
                yield post_process_fn(name, param, self._named_parameters)
        save_safetensors(
            _param_iterator(),
            checkpoint_path,
            total_len=len(self._named_parameters),
            is_io_worker=self.rank_in_shard_group == 0
        )

    @abort_if_not_cuda
    def load(self, checkpoint_path: str):
        state_dict_iterator = load_safetensors_generator(
            checkpoint_path,
            device="cuda",
            use_tqdm=self.rank_in_shard_group == 0,
            dtype=torch.bfloat16
        )
        self.load_state_dict(state_dict_iterator)

    # Private methods

    def _init_model(self, module: Fsdp2TrainModule, mesh_shape: tuple[int, ...]):
        device_mesh = self._init_device_mesh(mesh_shape)

        named_buffers = {
            name: buffer.cpu().clone()
            for name, buffer in module.named_buffers()
        }

        self.model = self._init_fsdp(module, device_mesh)

        for name, buffer in self.model.named_buffers():
            buffer.data.copy_(named_buffers[name])

        # Rebind the attributes
        for key, value in module.__dict__.items():
            if not key.startswith("_") and not hasattr(self.model, key):
                setattr(self.model, key, value)

        mat_src = module.materialization_source()
        if isinstance(mat_src, str):
            self.load(mat_src)
        elif isinstance(mat_src, dict):
            raise NotImplementedError("Currently not implemented, coming soon.")
        else:
            raise ValueError(f"Unknown materialization source: {mat_src}")

        # Call the on_weights_materialized hook if it exists
        if hasattr(self.module, "on_weights_materialized"):
            self.module.on_weights_materialized.__get__(self.model)()

    def _init_device_mesh(self, mesh_shape: tuple[int, ...]):
        from torch.distributed.device_mesh import init_device_mesh

        dist.init_process_group(backend="nccl")
        torch.cuda.set_device(0)
        device_mesh = init_device_mesh("cuda", mesh_shape, mesh_dim_names=("replicate", "shard"))
        return device_mesh

    def _init_fsdp(self, module: Fsdp2TrainModule, device_mesh: DeviceMesh):
        from torch.distributed.fsdp import MixedPrecisionPolicy, fully_shard

        mp_policy = MixedPrecisionPolicy(
            param_dtype=torch.bfloat16,
            reduce_dtype=torch.float32,
        )
        for sub_module in module.modules():
            if sub_module.__class__ in module.non_split_modules:
                fully_shard(sub_module, mesh=device_mesh, mp_policy=mp_policy)
        fully_shard(module, mesh=device_mesh, mp_policy=mp_policy)
        module.to_empty(device="cuda")
        module.init_weights()
        return module

    @cached_property
    def _named_parameters(self):
        return dict(self.model.named_parameters())

    def __del__(self):
        if dist.is_initialized():
            dist.destroy_process_group()



def flatten(item: list[list[any]]) -> list[any]:
    if not hasattr(item, "__iter__") or not hasattr(item, "__getitem__"):
        return item
    return sum([flatten(i) for i in item], [])


def set_random_seed(seed: int):
    import random

    import numpy
    import torch

    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)



import transformers


def patch_from_pretrained(cls):
    orig_from_pretrained = cls.from_pretrained

    def new_from_pretrained(cls, pretrained_model_name_or_path, *args, **kwargs):
        config = orig_from_pretrained(pretrained_model_name_or_path, *args, **kwargs)
        if isinstance(config, tuple):
            config[0].name_or_path = pretrained_model_name_or_path
        else:
            config.name_or_path = pretrained_model_name_or_path
        return config

    cls.from_pretrained = classmethod(new_from_pretrained)


def patch_all_configs():
    for subclass in transformers.PretrainedConfig.__subclasses__():
        patch_from_pretrained(subclass)


patch_all_configs()



from __future__ import annotations

import abc
from functools import wraps

import ray

from rlite.interface.base_executor import BaseExecutor
from rlite.resman import ResourceConsumer, ResourceManager


def abort_if_no_executor(func: callable):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.executors is None or len(self.executors) == 0:
            raise RuntimeError("Executor is not built")
        return func(self, *args, **kwargs)
    return wrapper


class BaseEngine(ResourceConsumer, abc.ABC):
    # ----- Interfaces ----- #

    @property
    @abc.abstractmethod
    def executor_cls(self) -> type[BaseExecutor]:
        raise NotImplementedError()

    # ----- Shared methods and properties ----- #

    def __init__(self, resource_manager: ResourceManager | None = None):
        if resource_manager is None:
            import rlite.resman
            resource_manager = rlite.resman.RESMAN
        ResourceConsumer.__init__(self, resource_manager)
        self._executors: list[BaseExecutor] = []

    @property
    def executors(self) -> list:
        return self._executors

    # Build methods

    def build(
        self,
        *,
        tensor_parallel_size: int,
        pipeline_parallel_size: int = 1,
        data_parallel_size: int = -1,
        allow_colocate: bool = True,
        colocate_with: ResourceConsumer | None = None,
        not_colocate_with: list[ResourceConsumer] | ResourceConsumer | None = None,
        ignore_rebuild_error: bool = False,
        **executor_kwargs
    ):
        if not self.is_registered():
            raise RuntimeError("Current engine is not registered in any resource manager!")

        if self.is_resource_allocated() and not ignore_rebuild_error:
            raise RuntimeError("Executors are already built")

        if data_parallel_size == -1:
            data_parallel_size = self._resource_manager.total_gpus // (
                tensor_parallel_size * pipeline_parallel_size
            )
            assert data_parallel_size > 0, "Not enough GPUs available!"

        # Ask for resources from resource manager
        num_gpu_request = tensor_parallel_size * pipeline_parallel_size * data_parallel_size
        self._resource_manager.allocate_resources(
            consumer=self,
            num_gpus=num_gpu_request,
            allow_colocate=allow_colocate,
            colocate_with=colocate_with,
            not_colocate_with=not_colocate_with,
        )

        # Allow some additional parameters to be added. this is useful, e.g.,
        # for FSDP, as different dp groups can also sync gradients (hybrid
        # sharding).
        additional_params = self.pre_build_executors_hook(
            tensor_parallel_size=tensor_parallel_size,
            pipeline_parallel_size=pipeline_parallel_size,
            data_parallel_size=data_parallel_size,
            allow_colocate=allow_colocate,
            colocate_with=colocate_with,
            not_colocate_with=not_colocate_with,
            ignore_rebuild_error=ignore_rebuild_error,
            **executor_kwargs
        )
        executor_kwargs.update(additional_params)

        # Assign resources to each executors (each executor is a dp group)
        executors = []
        for dp_rank in range(data_parallel_size):
            executor = self.build_executor(
                dp_rank=dp_rank,
                data_parallel_size=data_parallel_size,
                tensor_parallel_size=tensor_parallel_size,
                pipeline_parallel_size=pipeline_parallel_size,
                **executor_kwargs
            )
            executors.append(executor)
        self._executors = executors

        # Build workers
        futures = sum([executor.build_workers() for executor in self.executors], [])
        ray.get(futures)

    def build_executor(
        self,
        dp_rank: int,
        data_parallel_size: int,
        tensor_parallel_size: int,
        pipeline_parallel_size: int,
        **executor_kwargs
    ):
        num_gpus_per_dp = tensor_parallel_size * pipeline_parallel_size
        bundle_index_offset = dp_rank * num_gpus_per_dp
        bundles = self._resource_bundles[bundle_index_offset:bundle_index_offset + num_gpus_per_dp]
        executor = self.executor_cls(
            bundles=bundles,
            dp_rank=dp_rank,
            data_parallel_size=data_parallel_size,
            tensor_parallel_size=tensor_parallel_size,
            pipeline_parallel_size=pipeline_parallel_size,
            **executor_kwargs
        )
        return executor

    def pre_build_executors_hook(self, **kwargs) -> dict:
        return {}

    # State dict methods

    @abort_if_no_executor
    def load_state_dict(self, state_dict: dict):
        futures = sum([executor.load_state_dict(state_dict) for executor in self.executors], [])
        ray.get(futures)

    @abort_if_no_executor
    def state_dict(self) -> dict:
        # By default, the state dict is on CPU and gathered to rank 0
        return ray.get(self.executors[0].state_dict())[0]

    @abort_if_no_executor
    def p2p_weight_sync(self, another: BaseEngine):
        assert self.is_colocate_with(another), "Must be colocated to use p2p weight sync!"

        my_workers = sum([executor.workers for executor in self.executors], [])
        another_workers = sum([executor.workers for executor in another.executors], [])

        futures = [
            worker.sync_weights_to.remote(another_worker)
            for worker, another_worker in zip(my_workers, another_workers)
        ]
        ray.get(futures)

    # Device methods

    @abort_if_no_executor
    def cpu(self, *args, **kwargs):
        futures = sum([executor.cpu(*args, **kwargs) for executor in self.executors], [])
        ray.get(futures)

    @abort_if_no_executor
    def meta(self, *args, **kwargs):
        futures = sum([executor.meta(*args, **kwargs) for executor in self.executors], [])
        ray.get(futures)

    @abort_if_no_executor
    def cuda(self, *args, **kwargs):
        futures = sum([executor.cuda(*args, **kwargs) for executor in self.executors], [])
        ray.get(futures)

    @abort_if_no_executor
    def to(self, device: str, *args, **kwargs):
        futures = sum([executor.to(device, *args, **kwargs) for executor in self.executors], [])
        ray.get(futures)

    @abort_if_no_executor
    def device(self) -> list[str] | str:
        futures = sum([executor.device() for executor in self.executors], [])
        devices = ray.get(futures)
        if all(device == devices[0] for device in devices):
            return devices[0]
        else:
            return devices

    @abort_if_no_executor
    def release_memory_cache(self, cpu: bool = False, gpu: bool = True):
        futures = sum([executor.release_memory_cache(cpu, gpu) for executor in self.executors], [])
        ray.get(futures)



from __future__ import annotations

import random
import socket
import uuid
from collections import defaultdict
from functools import cached_property
from typing import TYPE_CHECKING, Literal

import ray
from ray.util.placement_group import PlacementGroup
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy

from rlite.resman.resource_bundle import ResourceBundle

if TYPE_CHECKING:
    from .resource_consumer import ResourceConsumer


class ResourceManager:
    """Resource manager that manages the resources of the cluster."""

    MAX_COLATE_COUNT = 20

    pg: PlacementGroup = None
    resource_bundles: list[ResourceBundle] = []
    node_resource_info: dict[str, dict[str, float]] = {}
    resource_allocations: dict[str, ResourceBundle] = {}
    allocated_ports: dict[str, list[int]] = {}

    def __init__(
        self,
        *,
        master_addr: str | None = None,
        network_interface: str = "eth0",
    ):
        if self.is_initialized():
            return

        self._network_interface = network_interface

        # Handle if there is no GPU in the cluster
        if "GPU" not in ray.cluster_resources() or ray.cluster_resources()["GPU"] == 0:
            self._gpu_available = False
        else:
            self._gpu_available = True

        # Record the node resource info
        ResourceManager.node_resource_info = {
            node["NodeID"]: {
                "CPU": int(node["Resources"].get("CPU", 0)),
                "GPU": int(node["Resources"].get("GPU", 0)),
                "IP": node["NodeManagerAddress"],
                "NodeID": node["NodeID"],
            }
            for node in ray.nodes()
        }

        # Create the resource bundles
        if self._gpu_available:
            for node in ResourceManager.node_resource_info.values():
                for _ in range(node["GPU"]):
                    ResourceManager.resource_bundles.append(
                        ResourceBundle(num_cpus=node["CPU"] // node["GPU"], num_gpus=1)
                    )
        else:
            for node in ResourceManager.node_resource_info.values():
                ResourceManager.resource_bundles.append(
                    ResourceBundle(num_cpus=node["CPU"], num_gpus=None)
                )

        # Set up the placement group
        self.pg = ray.util.placement_group(
            strategy="PACK",
            bundles=[
                bundle.to_dict(include_device_info=False)
                for bundle in ResourceManager.resource_bundles
            ],
        )
        ray.get(self.pg.ready())

        # Bind actual resources to the resource bundles
        mapping = ray.util.placement_group_table(self.pg)["bundles_to_node_id"]

        device_id_book = defaultdict(int)
        for index, bundle in enumerate(ResourceManager.resource_bundles):
            node_id = mapping[index]
            bundle.bind_resources(
                bundle_id=index,
                device_id=device_id_book[node_id],
                node_id=node_id,
                placement_group=self.pg,
                resource_manager=self,
            )
            device_id_book[node_id] += 1

        self._sort_resource_bundles(master_addr)

    def register(self, consumer: ResourceConsumer, consumer_id: str | None = None):
        if consumer_id is None:
            consumer_id = str(uuid.uuid4())[:16]

        if hasattr(consumer, "_consumer_id") and consumer._consumer_id is not None:
            if consumer._consumer_id != consumer_id:
                raise ValueError("The consumer_id is already registered!")
            return

        # Initialize the consumer
        consumer._resource_manager = self
        consumer._consumer_id = consumer_id
        consumer._resource_bundles = []
        self.resource_allocations[consumer_id] = []

    def unregister(self, consumer: ResourceConsumer):
        for bundle in self.resource_allocations.pop(consumer._consumer_id):
            self.resource_bundles[bundle.bundle_id_in_placement_group] += bundle

        delattr(consumer, "_resource_manager")
        delattr(consumer, "_consumer_id")
        delattr(consumer, "_resource_bundles")

    def allocate_resources(
        self,
        consumer: ResourceConsumer,
        num_gpus: int | None = None,
        colocate_with: ResourceConsumer | None = None,
        not_colocate_with: list[ResourceConsumer] | ResourceConsumer | None = None,
        allow_colocate: bool = False,
    ) -> bool:
        if not self.is_initialized():
            raise ValueError("Resource manager is not initialized!")

        if not_colocate_with is not None:
            if not isinstance(not_colocate_with, list):
                not_colocate_with = [not_colocate_with]
            forbidden_bundle_ids = set([
                bundle.bundle_id_in_placement_group
                for consumer in not_colocate_with
                for bundle in consumer._resource_bundles
            ])
        else:
            forbidden_bundle_ids = set()

        # Colocate with the specified consumer
        if colocate_with is not None:
            # Sanity check
            if not colocate_with.is_registered(manager=self):
                raise ValueError("The specified consumer has not been registered!")
            if not colocate_with.is_resource_allocated():
                raise ValueError(
                    "The specified colocate_with consumer has not been allocated resources!"
                )
            bundles = self._get_bundles_from_colocate_consumer(colocate_with, forbidden_bundle_ids)
        else:
            # TODO: handle requires no gpu
            if num_gpus is None:
                raise ValueError("num_gpus is required when colocate_with is not specified!")

            if num_gpus > self.total_gpus:
                raise ValueError(f"Not enough GPUs available! {num_gpus} > {self.total_gpus}")

            if num_gpus > self.available_resource_bundles("gpu"):
                raise ValueError(
                    f"Not enough GPUs available! "
                    f"{num_gpus} > {self.available_resource_bundles('gpu')}"
                )

            bundles = self._get_bundles_from_gpu_request(
                gpu_request=num_gpus,
                allow_colocate=allow_colocate,
                forbidden_bundle_ids=forbidden_bundle_ids,
            )

        # Assign bundles to the consumer
        for bundle in bundles:
            self.resource_allocations[consumer._consumer_id].append(bundle)
            consumer._resource_bundles.append(bundle)

    def _get_bundles_from_colocate_consumer(
        self,
        colocate_with: ResourceConsumer,
        forbidden_bundle_ids: set[int],
    ) -> list[ResourceBundle]:
        # Ensure the bundles not overlap with the not_colocate_with's bundles
        bundle_ids_to_allocate = set([
            bundle.bundle_id_in_placement_group
            for bundle in colocate_with._resource_bundles
        ])
        assert len(forbidden_bundle_ids & bundle_ids_to_allocate) == 0, (
            "The specified colocate_with consumer has overlapping bundles "
            "with the forbidden bundles!"
        )

        # Use the colocate_with's resource bundles
        bundles = []
        for bundle in colocate_with._resource_bundles:
            new_bundle = ResourceBundle(
                num_cpus=self.colocate_num_cpus,
                num_gpus=self.colocate_num_gpus
            )
            new_bundle.bind_resources(
                bundle_id=bundle.bundle_id_in_placement_group,
                device_id=bundle.device_id,
                node_id=bundle.node_id,
                placement_group=bundle.placement_group,
                resource_manager=self,
            )
            source_bundle = self.get_bundle_by_id(bundle.bundle_id_in_placement_group)
            source_bundle -= new_bundle
            bundles.append(new_bundle)

        return bundles

    def _get_bundles_from_gpu_request(
        self,
        gpu_request: int,
        allow_colocate: bool,
        forbidden_bundle_ids: set[int],
    ) -> list[ResourceBundle]:
        # NOTE: as one bundle can hold at most 1 GPU, we need to allocate
        #       num_gpus bundles
        num_cpus = self.colocate_num_cpus if allow_colocate else 1
        num_gpus = self.colocate_num_gpus if allow_colocate else 1

        # Find the available bundles
        bundles_to_allocate = []
        target_bundle = ResourceBundle(num_cpus=num_cpus, num_gpus=num_gpus)
        for bundle in self.resource_bundles:
            if len(bundles_to_allocate) == gpu_request:
                break
            if bundle.bundle_id_in_placement_group in forbidden_bundle_ids:
                continue
            if bundle >= target_bundle:
                bundles_to_allocate.append(bundle)

        if len(bundles_to_allocate) < gpu_request:
            raise ValueError(
                f"Not enough gpus to allocate! {len(bundles_to_allocate)} < {gpu_request}"
            )

        # Allocate resources
        bundles = []
        for bundle in bundles_to_allocate:
            new_bundle = ResourceBundle(num_cpus=num_cpus, num_gpus=num_gpus)
            new_bundle.bind_resources(
                bundle_id=bundle.bundle_id_in_placement_group,
                device_id=bundle.device_id,
                node_id=bundle.node_id,
                placement_group=bundle.placement_group,
                resource_manager=self,
            )
            bundle -= new_bundle
            bundles.append(new_bundle)

        return bundles

    def deallocate_resources(self, consumer: any, bundles: list[ResourceBundle] | None = None):
        if bundles is not None:
            bundle_ids_to_deallocate = [bundle.bundle_id_in_placement_group for bundle in bundles]
        else:
            bundle_ids_to_deallocate = [
                bundle.bundle_id_in_placement_group
                for bundle in self.resource_allocations[consumer._consumer_id]
            ]

        bundles_to_keep = []
        for bundle in self.resource_allocations[consumer._consumer_id]:
            if bundle.bundle_id_in_placement_group not in bundle_ids_to_deallocate:
                bundles_to_keep.append(bundle)
            else:
                # Reclaim the resources
                self.resource_bundles[bundle.bundle_id_in_placement_group] += bundle
        self.resource_allocations[consumer._consumer_id] = bundles_to_keep
        consumer._resource_bundles = bundles_to_keep

    def available_resource_bundles(
        self, resource_type: Literal["gpu", "cpu"] = "gpu", allow_colocate: bool = True
    ) -> int:
        result = 0
        for bundle in self.resource_bundles:
            if resource_type == "gpu" and bundle.num_gpus is not None:
                threshold = self.colocate_num_gpus if allow_colocate else 1
                if bundle.num_gpus >= threshold:
                    result += 1
            if resource_type == "cpu" and bundle.num_cpus is not None:
                threshold = self.colocate_num_cpus if allow_colocate else 1
                if bundle.num_cpus >= threshold:
                    result += 1
        return result

    def get_free_port(self, bundle: ResourceBundle | int) -> int:
        """Get a free port from the host node of the bundle."""

        if isinstance(bundle, int):
            bundle = self.resource_bundles[bundle]

        def test_port_available(port: int) -> bool:
            def _test_port_available(port: int) -> bool:
                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                    return s.connect_ex(("localhost", port)) != 0

            return ray.get(
                ray.remote(_test_port_available)
                .options(
                    scheduling_strategy=PlacementGroupSchedulingStrategy(
                        placement_group=self.pg,
                        placement_group_bundle_index=bundle.bundle_id_in_placement_group,
                    )
                )
                .remote(port)
            )

        def port_allocated(port: int) -> bool:
            return port in self.allocated_ports.get(bundle.node_id, [])

        if not self.is_initialized():
            raise ValueError("Resource manager is not initialized!")

        port = random.randint(1024, 65535)
        while port_allocated(port) or not test_port_available(port):
            port = random.randint(1024, 65535)

        self.allocated_ports.setdefault(bundle.node_id, []).append(port)

        return port

    def get_ip_addr(self, bundle: ResourceBundle | int) -> str:
        """Get the IP address of the host node of the given bundle."""

        if isinstance(bundle, int):
            bundle = self.resource_bundles[bundle]
        return ResourceManager.node_resource_info[bundle.node_id]["IP"]

    @cached_property
    def colocate_num_gpus(self) -> float:
        return 1 / self.MAX_COLATE_COUNT

    @property
    def colocate_num_cpus(self) -> float:
        return 1.0 / self.MAX_COLATE_COUNT

    @cached_property
    def total_gpus(self) -> float:
        return sum(
            bundle.num_gpus
            for bundle in self.resource_bundles
            if bundle.num_gpus is not None
        )

    def is_initialized(self) -> bool:
        return self.pg is not None

    def get_bundle_by_id(self, bundle_id: int):
        """Fetch the bundle by id, useful after sorting the bundles."""
        for bundle in self.resource_bundles:
            if bundle.bundle_id_in_placement_group == bundle_id:
                return bundle
        raise ValueError(f"Bundle ID {bundle_id} not found!")

    def _sort_resource_bundles(self, master_addr: str | None = None):
        # NOTE (zhanghan): We need to sort the resource bundles by the node id,
        #                  so that the head node will always be the first one,
        #                  and the resource bundles on the same node will be
        #                  grouped together.
        def _get_ip_addr_by_network_interface(network_interface: str) -> str:
            import psutil
            for interface, addrs in psutil.net_if_addrs().items():
                for addr in addrs:
                    if addr.family == 2 and interface == network_interface:
                        return addr.address
            raise ValueError(f"Network interface {network_interface} not found!")

        if master_addr is None:
            try:
                master_addr = _get_ip_addr_by_network_interface(self._network_interface)
            except ValueError:
                master_addr = socket.gethostbyname(socket.gethostname())
        for node in self.node_resource_info.values():
            if node["IP"] == master_addr:
                head_node_id = node["NodeID"]
                break
        ResourceManager.resource_bundles = list(
            sorted(
                ResourceManager.resource_bundles,
                key=lambda x: (x.node_id != head_node_id, x.node_id)
            )
        )

    def __del__(self):
        if ray.is_initialized():
            ray.shutdown()


RESMAN: ResourceManager | None = None



import abc

from rlite.utils.distributed import SerializableModule


class BaseInferenceModule(SerializableModule, abc.ABC):
    @abc.abstractmethod
    def get_checkpoint_path(self) -> str:
        """Get the checkpoint path."""
        raise NotImplementedError()



import abc

from rlite.interface import BaseWorker


class BaseTrainWorker(BaseWorker, abc.ABC):
    @abc.abstractmethod
    def save(self, checkpoint_path: str):
        raise NotImplementedError()

    @abc.abstractmethod
    def load(self, checkpoint_path: str):
        raise NotImplementedError()

    def getattr(self, name, *args, **kwargs):
        attr = getattr(self, name)
        if callable(attr):
            return attr(*args, **kwargs)
        return attr

    def hasattr(self, name):
        return hasattr(self, name) and callable(getattr(self, name))



import subprocess

import rlite
from rlite import RliteInferenceEngine

if __name__ == "__main__":
    rlite.init()

    model_name_or_path = "Qwen/Qwen2.5-7B-Instruct"
    engine = RliteInferenceEngine(model_name_or_path, executor="vllm")
    engine.build(tensor_parallel_size=4)  # On GPU

    engine.release_memory_cache(gpu=True, cpu=False)

    print(f"After building, device: \033[92m{engine.device()}\033[0m")

    subprocess.run(["nvidia-smi"])

    engine.cpu()  # On CPU.

    print(f"After cpu(), device: \033[92m{engine.device()}\033[0m")

    subprocess.run(["nvidia-smi"])

    engine.cuda()  # On GPU again.

    print(f"After cuda(), device: \033[92m{engine.device()}\033[0m")

    subprocess.run(["nvidia-smi"])

    engine.meta()

    print(f"After meta(), device: \033[92m{engine.device()}\033[0m")

    subprocess.run(["nvidia-smi"])



from __future__ import annotations

import json
import os
import warnings
from datetime import timedelta
from pathlib import Path
from typing import Any, Generator, Optional, Union

import cloudpickle
import pynvml
import torch
import torch.distributed
from safetensors import safe_open
from safetensors.torch import save_file
from torch.distributed.distributed_c10d import (
    Backend,
    GroupMember,
    PrefixStore,
    ProcessGroup,
    Store,
    _new_process_group_helper,
    _shutdown_backend,
    _unregister_all_process_groups,
    _unregister_process_group,
    _update_default_pg,
    _world,
    default_pg_timeout,
    rendezvous
)
from tqdm import tqdm


class CUDAIPCHandle:
    def __init__(
        self,
        tensor_type: type,
        size: tuple,
        stride: tuple,
        offset: int,
        storage_type: type,
        dtype: torch.dtype,
        device: torch.device,
        handle: bytes,
        storage_size_bytes: bytes,
        storage_offset_bytes: bytes,
        requires_grad: bool,
        ref_counter_handle: bytes,
        ref_counter_offset: bytes,
        event_handle: bytes,
        event_sync_required: bool,
    ):
        self.tensor_type = tensor_type
        self.size = size
        self.stride = stride
        self.offset = offset
        self.storage_type = storage_type
        self.dtype = dtype
        self.device = device
        self.handle = handle
        self.storage_size_bytes = storage_size_bytes
        self.storage_offset_bytes = storage_offset_bytes
        self.requires_grad = requires_grad
        self.ref_counter_handle = ref_counter_handle
        self.ref_counter_offset = ref_counter_offset
        self.event_handle = event_handle
        self.event_sync_required = event_sync_required

    def rebuild(self) -> torch.Tensor:
        # NOTE: Rebuild within the same process is not thread-safe and will
        #       likely crash (segfault core dump).
        if self.handle is None:
            raise ValueError("Storage handle is None")
        torch.cuda._lazy_init()
        storage = self.storage_type._new_shared_cuda(
            self.device,
            self.handle,
            self.storage_size_bytes,
            self.storage_offset_bytes,
            self.ref_counter_handle,
            self.ref_counter_offset,
            self.event_handle,
            self.event_sync_required,
        )
        _storage = (
            storage
            if isinstance(storage, torch.UntypedStorage)
            else storage._untyped_storage
        )

        self.storage_type._release_ipc_counter(
            self.ref_counter_handle,
            self.ref_counter_offset,
            device=self.device
        )

        t = torch._utils._rebuild_tensor(
            torch.storage.TypedStorage(wrap_storage=_storage, dtype=self.dtype, _internal=True),
            self.offset,
            self.size,
            self.stride,
        )

        if self.tensor_type == torch.nn.parameter.Parameter:
            # It is crucial for integer tensors to receive
            # the requires_grad=False as an argument in the constructor
            t = torch.nn.parameter.Parameter(t, requires_grad=self.requires_grad)
        else:
            t.requires_grad = self.requires_grad

        return t.clone()

    @staticmethod
    def from_tensor(tensor: torch.Tensor) -> CUDAIPCHandle:
        if not tensor.is_cuda:
            raise ValueError("Tensor must be on CUDA device to use CUDAIPC")
        tensor = tensor.clone().share_memory_()
        storage = tensor._typed_storage()
        (
            device,
            handle,
            storage_size_bytes,
            storage_offset_bytes,
            ref_counter_handle,
            ref_counter_offset,
            event_handle,
            event_sync_required,
        ) = storage._share_cuda_()
        tensor_offset = tensor.storage_offset()
        storage._release_ipc_counter(ref_counter_handle, ref_counter_offset, device=device)
        return CUDAIPCHandle(
            tensor_type=type(tensor),
            size=tensor.size(),
            stride=tensor.stride(),
            # tensor offset in its storage
            offset=tensor_offset,
            storage_type=type(storage),
            dtype=tensor.dtype,
            device=device,
            # identifier which CUDA allocation is the storage in.
            handle=handle,
            # size(in bytes) of the storage
            storage_size_bytes=storage_size_bytes,
            # offset(in bytes) of the storage in the CUDA allocation
            storage_offset_bytes=storage_offset_bytes,
            requires_grad=tensor.requires_grad,
            ref_counter_handle=ref_counter_handle,
            ref_counter_offset=ref_counter_offset,
            event_handle=event_handle,
            event_sync_required=event_sync_required,
        )


def get_device_uuid(device_id: int = 0) -> str:
    pynvml.nvmlInit()
    try:
        if "CUDA_VISIBLE_DEVICES" in os.environ:
            device_ids = os.environ["CUDA_VISIBLE_DEVICES"].split(",")
            if device_ids == [""]:
                raise RuntimeError(
                    "CUDA_VISIBLE_DEVICES is set to empty string,"
                    " which means GPU support is disabled."
                )
            physical_device_id = int(device_ids[device_id])
        else:
            physical_device_id = device_id

        handle = pynvml.nvmlDeviceGetHandleByIndex(physical_device_id)
        return pynvml.nvmlDeviceGetUUID(handle)
    finally:
        pynvml.nvmlShutdown()


def load_safetensors_generator(
    safetensors_path,
    device="cpu",
    use_tqdm: bool = True,
    fetch_data: bool = True,
    dtype: torch.dtype = torch.float32,
):
    path = Path(safetensors_path)
    if not path.exists():
        raise FileNotFoundError(f"File not found: {safetensors_path}")

    if path.is_dir():
        paths = list(path.glob("*.safetensors"))
        if len(paths) == 0:
            raise FileNotFoundError(f"No safetensors files found in directory: {safetensors_path}")
    else:
        paths = [path]

    if use_tqdm:
        paths = tqdm(paths, desc="Loading safetensors")

    for safetensors_path in paths:
        with safe_open(safetensors_path, framework="pt", device=device) as f:
            for key in f.keys():
                if fetch_data:
                    yield key, f.get_tensor(key).to(dtype)
                else:
                    yield key, None


def save_safetensors(
    state_dict: dict | Generator,
    path: str,
    min_block_size_in_GB: int = 5,
    max_block_size_in_GB: int = 20,
    total_size: int | None = None,
    total_len: int | None = None,
    is_io_worker: bool = True,
):
    path = Path(path)
    path.mkdir(parents=True, exist_ok=True)

    if total_size is None and isinstance(state_dict, dict):
        total_size = 0
        for param in state_dict.values():
            total_size += param.numel() * param.element_size()

    if total_size is not None:
        size_per_block = total_size // 4 + 1
        size_per_block = min(size_per_block, max_block_size_in_GB * 1024**3)
        size_per_block = max(size_per_block, min_block_size_in_GB * 1024**3)
    else:
        from loguru import logger

        logger.warning(
            f"Total size is not provided, using min block size: {min_block_size_in_GB}GB."
        )
        size_per_block = min_block_size_in_GB * 1024**3

    if total_len is None and isinstance(state_dict, dict):
        total_len = len(state_dict)

    if isinstance(state_dict, dict):
        state_dict = state_dict.items()

    partial_state_dict = {}
    num_saved_files = 0
    partial_state_dict_size = 0
    safetensors_path = path / "model-1.safetensors"
    meta_info = {"metadata": {"total_size": total_size}, "weight_map": {}}

    for name, param in tqdm(
        state_dict,
        total=total_len,
        desc="Saving weights",
        disable=not is_io_worker,
    ):
        partial_state_dict[name] = param
        partial_state_dict_size += param.numel() * param.element_size()
        meta_info["weight_map"][name] = safetensors_path.name
        if partial_state_dict_size > size_per_block:
            if is_io_worker:
                save_file(partial_state_dict, safetensors_path, metadata={"format": "pt"})
            num_saved_files += 1
            safetensors_path = path / f"model-{num_saved_files + 1}.safetensors"

            partial_state_dict = {}
            partial_state_dict_size = 0

    if len(partial_state_dict) > 0 and is_io_worker:
        save_file(partial_state_dict, safetensors_path, metadata={"format": "pt"})

    if is_io_worker:
        with open(path / "model.safetensors.index.json", "w") as f:
            json.dump(meta_info, f)


def sync_weights(
    source_model,
    target_model,
    source_model_reside: str = "cpu",
    target_model_reside: str = "cuda"
):
    """Synchronize the weights from the train model to the inference model.

    This should happen when train model is on GPU and inference model is on CPU
    or Meta.
    """
    # TODO (zhanghan):
    # we can further reduce the GPU memory usage by lazy loading the weights.
    # That is, during iterating the generator, we gradually materialize the
    # inference model's weights on GPU. In the meantime, we can free the
    # weights from the train model. This allows that at any time, only 1x of
    # the model size is on GPU. Currently, we always keep 2x of the model size
    # on GPU.
    torch.cuda.empty_cache()
    source_model.clear_cache()
    target_model.clear_cache()

    if source_model.device != "cuda":
        source_model.cuda(init_cache_engine=False)
    if target_model.device != "cuda":
        target_model.cuda(init_cache_engine=False)
    for model in (source_model, target_model):
        if hasattr(model, "is_cache_engine_active") and model.is_cache_engine_active():
            model.free_cache_engine()

    if source_model.is_colocate_with(target_model):
        # P2P weight sync via CUDA IPC
        source_model.sync_weights_cudaipc(target_model)
    else:
        # All-gather, shared CPU memory, and broadcast
        target_model.load_state_dict(source_model.state_dict_generator())

    torch.cuda.empty_cache()
    source_model.to(source_model_reside)
    target_model.to(target_model_reside)


class SerializableModule(torch.nn.Module):
    def __getstate__(self):
        return cloudpickle.dumps(self.__dict__)

    def __setstate__(self, state):
        self.__dict__.update(cloudpickle.loads(state))


# Copy from pytorch to allow creating multiple main groups.
# https://github.com/pytorch/pytorch/blob/main/torch/distributed/distributed_c10d.py
def init_process_group(
    backend: Union[str, Backend] = None,
    init_method: Optional[str] = None,
    timeout: Optional[timedelta] = None,
    world_size: int = -1,
    rank: int = -1,
    store: Optional[Store] = None,
    group_name: str = None,
    pg_options: Optional[Any] = None,
):
    assert (store is None) or (init_method is None), "Cannot specify both init_method and store."

    if store is not None:
        assert world_size > 0, "world_size must be positive if using store"
        assert rank >= 0, "rank must be non-negative if using store"
    elif init_method is None:
        init_method = "env://"

    if backend:
        backend = Backend(backend)
    else:
        backend = Backend("undefined")

    if timeout is None:
        timeout = default_pg_timeout

    # backward compatible API
    if store is None:
        rendezvous_iterator = rendezvous(init_method, rank, world_size, timeout=timeout)
        store, rank, world_size = next(rendezvous_iterator)
        store.set_timeout(timeout)

        # Use a PrefixStore to avoid accidental overrides of keys used by
        # different systems (e.g. RPC) in case the store is multi-tenant.
        store = PrefixStore(group_name, store)

    pg_options_param_name = "backend_options" if str(torch.__version__) >= "2.6" else "pg_options"
    pg, _ = _new_process_group_helper(
        world_size,
        rank,
        [],
        backend,
        store,
        group_name=group_name,
        **{pg_options_param_name: pg_options},
        timeout=timeout,
    )

    _world.pg_group_ranks[pg] = {i: i for i in range(world_size)}

    return pg


def destroy_process_group(group: Optional[ProcessGroup] = None):
    """
    Destroy a given process group, and deinitialize the distributed package.

    Args:
        group (ProcessGroup, optional): The process group to be destroyed, if
                                        group.WORLD is given, all process
                                        groups including the default one will
                                        be destroyed.
    """
    global _world

    if group == GroupMember.NON_GROUP_MEMBER:
        return

    if group is None:
        pg = GroupMember.WORLD
    else:
        pg = group

    assert pg is not None
    if _world.pg_map.get(pg, None) is None:
        raise ValueError("Invalid process group specified")

    # When users register Python onCompletion hooks, those hooks will run on a
    # different thread than the main thread. Today, the ProcessGroup dtor does
    # wait for that thread. However, the dtor might finish after the Python
    # Interpreter exits. After that grabbing the GIL for the Python hook will
    # crash. We can either revive the interpreter when running hooks or keep
    # the main one alive until all works and hooks are done. The current
    # implementation does the latter. Therefore, we explicitly call
    # _wait_for_pending_works() here to wait for the pending hooks to finish.
    if pg.name().lower() == "nccl" and pg._has_hooks():
        pg._wait_for_pending_works()

    if group is None or group == GroupMember.WORLD:
        # shutdown all backends in the order of pg names. shutting down in
        # order because ncclCommAbort() was a 'collective' call in some
        # versions of NCCL.
        for pg_to_shutdown in sorted(
            _world.pg_names,
            key=lambda x: _world.pg_names[x],
            reverse=True
        ):
            _shutdown_backend(pg_to_shutdown)

        _update_default_pg(None)
        _world.pg_map.clear()
        _world.pg_names.clear()
        _world.pg_group_ranks.clear()
        _world.pg_backend_config.clear()
        _world.pg_to_tag.clear()
        _world.tags_to_pg.clear()
        _world.pg_coalesce_state.clear()
        _world.pg_default_device.clear()
        _unregister_all_process_groups()

        # when process group doesn't have an explicit name (only WORLD
        # (default) process group can have an explicit name), we use global
        # _world.group_count to generate the name. We need to reset the counter
        # on destruction to allow consistent value to be generated when we
        # re-create process groups after some trainers recover from failure.
        #
        # We only reset this when WORLD is being destroyed because if this
        # process group is in good state, we aren't dealing with failures.
        _world.group_count = 0
    else:
        _shutdown_backend(pg)
        del _world.pg_map[pg]
        del _world.pg_names[pg]
        del _world.pg_group_ranks[pg]
        del _world.pg_backend_config[pg]
        if pg in _world.pg_default_device:
            del _world.pg_default_device[pg]
        if pg in _world.pg_coalesce_state.keys():
            warnings.warn(
                "Some coalesced collectives haven't been launched when "
                "ProcessGroup is destroyed. They will be cleaned."
            )
            del _world.pg_coalesce_state[pg]

        tag = _world.pg_to_tag.get(pg)
        del _world.pg_to_tag[pg]
        if tag is not None:
            try:
                _world.tags_to_pg[tag].remove(pg)
                if tag.startswith("ptd:"):
                    _world.tags_to_pg[""].remove(pg)
            except Exception:
                pass
        _unregister_process_group(pg.group_name)



import rlite
from rlite import RliteInferenceEngine

if __name__ == "__main__":
    rlite.init()

    model_name_or_path = "Qwen/Qwen2.5-7B-Instruct"
    prompts = [
        "你好，请介绍一下你自己。",
        "Hello, how are you?",
        "中国的首都是",
        "France is famous for its",
    ]

    engine = RliteInferenceEngine(model_name_or_path, executor="vllm")
    engine.build(tensor_parallel_size=4)

    sampling_params = {"temperature": 0.7, "seed": 42}
    outputs = engine.generate(prompts, sampling_params=sampling_params)

    for prompt, output in zip(prompts, outputs):
        print(f"\033[92m{prompt}\033[94m{output.outputs[0].text}\033[0m")



import abc

from rlite.interface.train.train_executor import (
    TRAIN_EXECUTOR_REGISTRY,
    BaseTrainExecutor
)
from rlite.nn import Fsdp2TrainModule
from rlite.resman import ResourceBundle
from rlite.train.fsdp2.worker import Fsdp2Worker
from rlite.utils.misc import set_random_seed


@TRAIN_EXECUTOR_REGISTRY.register([
    "fsdp2", "FSDP2", "Fsdp2", "Fsdp2TrainExecutor"
])
class Fsdp2TrainExecutor(BaseTrainExecutor, abc.ABC):
    def __init__(
        self,
        *,
        dp_rank: int,
        module: Fsdp2TrainModule,
        bundles: list[ResourceBundle],
        tensor_parallel_size: int,
        data_parallel_size: int = 1,
        pipeline_parallel_size: int = 1,
        seed: int = 42,
        master_addr: str,
        master_port: int,
        mesh_shape: tuple[int, ...],
        **fsdp_kwargs
    ):
        # assert isinstance(module, Fsdp2TrainModule), type(module)
        self.dp_rank = dp_rank
        self.module = module
        self.bundles = bundles
        self.tensor_parallel_size = tensor_parallel_size
        self.data_parallel_size = data_parallel_size
        self.pipeline_parallel_size = pipeline_parallel_size
        self.seed = seed
        self.fsdp_kwargs = fsdp_kwargs
        self.master_addr = master_addr
        self.master_port = master_port
        self.mesh_shape = mesh_shape

        self._workers: list[Fsdp2Worker] = []

    @property
    def workers(self):
        return self._workers

    def build_workers(self):
        self._workers = self._build_workers(
            module=self.module,
            bundles=self.bundles,
            master_addr=self.master_addr,
            master_port=self.master_port,
            mesh_shape=self.mesh_shape,
            **self.fsdp_kwargs
        )
        return [worker.ready.remote() for worker in self._workers]

    def _build_workers(
        self,
        module: Fsdp2TrainModule,
        bundles: list[ResourceBundle],
        master_addr: str,
        master_port: int,
        mesh_shape: tuple[int, ...],
        **fsdp_kwargs
    ):
        set_random_seed(self.seed)
        assert len(bundles) == self.tensor_parallel_size * self.pipeline_parallel_size, (
            f"The number of bundles ({len(bundles)=}) "
            "should be equal to the product of "
            f"tensor_parallel_size ({self.tensor_parallel_size=}) "
            f"and pipeline_parallel_size ({self.pipeline_parallel_size=})"
        )

        assert self.pipeline_parallel_size == 1, "Pipeline parallel is not supported yet!"
        assert self.tensor_parallel_size > 1, "FSDP2 requires tensor parallel size > 1"

        workers = []
        for rank in range(self.tensor_parallel_size * self.pipeline_parallel_size):
            bundle = bundles[rank]
            global_rank = (
                self.dp_rank * self.tensor_parallel_size * self.pipeline_parallel_size + rank
            )
            global_world_size = (
                self.data_parallel_size * self.tensor_parallel_size * self.pipeline_parallel_size
            )
            worker = bundle.remote(
                Fsdp2Worker,
                runtime_env={"env_vars": {
                    "RANK": str(global_rank),
                    "WORLD_SIZE": str(global_world_size),
                    "MASTER_ADDR": master_addr,
                    "MASTER_PORT": str(master_port),
                }}
            ).remote(
                module=module,
                mesh_shape=mesh_shape,
                rank_in_shard_group=rank,
                shard_group_world_size=self.tensor_parallel_size * self.pipeline_parallel_size,
                **fsdp_kwargs
            )
            workers.append(worker)
        return workers



import transformers
from accelerate import init_empty_weights
from transformers import AutoConfig

import rlite
import rlite.nn
from rlite import RliteTrainEngine

if __name__ == "__main__":
    rlite.init()

    config = AutoConfig.from_pretrained(
        "Qwen/Qwen2.5-7B-Instruct",
        attn_implementation="flash_attention_2"
    )
    actor_cls = getattr(transformers, config.architectures[0])

    # Construct a class based on the config
    class Actor(actor_cls, rlite.nn.HuggingFaceFsdp2TrainModule):
        ...

    with init_empty_weights():
        module = Actor(config)
    engine = RliteTrainEngine(module, executor="fsdp2")
    engine.build(tensor_parallel_size=4)



import abc

import ray

from rlite.interface.base_executor import BaseExecutor
from rlite.train.utils import prepare_parallel_args, prepare_parallel_kwargs
from rlite.utils.registry import Registry


class BaseTrainExecutor(BaseExecutor, abc.ABC):
    def run_workers(self, method_name: str, *args, **kwargs) -> list[ray.ObjectRef]:
        """Run a method on all workers.

        This function allows calling the methods defined in the module. The
        args and kwargs will be parsed according to its NeedParallel type.
        """
        # Magic part.
        args_list = prepare_parallel_args(args, unwrap=True, dp_size=len(self.workers))
        kwargs_list = prepare_parallel_kwargs(kwargs, unwrap=True, dp_size=len(self.workers))
        if hasattr(self.workers[0], method_name):
            futures = [
                getattr(worker, method_name).remote(*args, **kwargs)
                for worker, args, kwargs in zip(self.workers, args_list, kwargs_list)
            ]
        else:
            futures = [
                worker.getattr.remote(method_name, *args, **kwargs)
                for worker, args, kwargs in zip(self.workers, args_list, kwargs_list)
            ]
        return futures

    def train(self) -> list[ray.ObjectRef]:
        """Switch to training mode."""
        return [worker.train.remote() for worker in self.workers]

    def eval(self) -> list[ray.ObjectRef]:
        """Switch to evaluation mode."""
        return [worker.eval.remote() for worker in self.workers]

    def is_training(self) -> list[bool]:
        return [worker.is_training() for worker in self.workers]

    def save(self, checkpoint_path: str, *args, **kwargs) -> list[ray.ObjectRef]:
        return [worker.save.remote(checkpoint_path, *args, **kwargs) for worker in self.workers]

    def load(self, checkpoint_path: str, *args, **kwargs) -> list[ray.ObjectRef]:
        return [worker.load.remote(checkpoint_path, *args, **kwargs) for worker in self.workers]


TRAIN_EXECUTOR_REGISTRY = Registry("Train Executors")



import torch
import torch.nn.functional as F
import torch.optim
from accelerate import init_empty_weights
from transformers import Qwen2Config, Qwen2ForCausalLM

import rlite
import rlite.nn
from rlite import NeedParallel, RliteTrainEngine


class MyQwenModel(rlite.nn.HuggingFaceFsdp2TrainModule, Qwen2ForCausalLM):
    def on_weights_materialized(self):
        self.optimizer = torch.optim.AdamW(self.parameters(), lr=1e-5)

    def train_step(self, input_ids: list[list[int]], gradient_accumulation_steps: int = 1):
        input_ids = torch.tensor(input_ids, dtype=torch.long, device="cuda")

        logits = self(input_ids).logits[:, :-1].contiguous()
        labels = input_ids[:, 1:].contiguous()

        loss = F.cross_entropy(logits.view(-1, logits.size(-1)), labels.view(-1)).mean()
        loss = loss / gradient_accumulation_steps
        loss.backward()

        return loss.item()

    def optimizer_step(self):
        self.optimizer.step()
        self.optimizer.zero_grad()


if __name__ == "__main__":
    rlite.init()

    with init_empty_weights():  # Init on meta device
        config = Qwen2Config.from_pretrained(
            "Qwen/Qwen2.5-1.5B-Instruct",
            attn_implementation="flash_attention_2"
        )
        # NOTE: MyQwenModel.from_pretrained may read the checkpoint files
        module = MyQwenModel(config)
        module.gradient_checkpointing_enable()

    engine = RliteTrainEngine(module, executor="fsdp2")  # Must on meta device
    # NOTE: here tensor_parallel_size means the number of parameter shards
    engine.build(tensor_parallel_size=4)

    num_gpus = 8
    gradient_accumulation_steps = 4

    losses = []

    for _ in range(gradient_accumulation_steps):
        input_ids = torch.randint(0, 10000, (num_gpus, 1024))
        losses_for_each_gpu = engine.train_step(
            NeedParallel(input_ids),
            gradient_accumulation_steps=gradient_accumulation_steps
        )
        losses.append(losses_for_each_gpu)

    print(losses)



from __future__ import annotations

import abc

from vllm import RequestOutput, SamplingParams

from rlite.interface import BaseExecutor
from rlite.utils.registry import Registry


class BaseInferenceExecutor(BaseExecutor, abc.ABC):
    @abc.abstractmethod
    def generate(
        self,
        prompts: list[str],
        sampling_params: dict | SamplingParams | None = None,
        use_tqdm: bool = True,
        tqdm_desc: str = "Processed prompts",
        return_none: bool = False,
    ) -> list[RequestOutput | None]:
        """Generate the output for the given prompts."""
        futures = [
            worker.generate.remote(
                prompts=prompts,
                sampling_params=sampling_params,
                use_tqdm=use_tqdm,
                tqdm_desc=tqdm_desc,
                return_none=(rank != 0)
            )
            for rank, worker in enumerate(self.workers)
        ]
        return futures

    def device(self) -> str:
        futures = [worker.get_device.remote() for worker in self.workers]
        return futures


INFERENCE_EXECUTOR_REGISTRY = Registry("Inference Executors")



import torch
from accelerate import init_empty_weights
from transformers import (
    AutoConfig,
    AutoProcessor,
    Qwen2_5_VLForConditionalGeneration
)

import rlite
import rlite.nn
from rlite import NeedParallel, RliteTrainEngine


def prepare_image(image_url: str):
    from io import BytesIO

    import requests
    from PIL import Image

    response = requests.get(image_url)
    image = Image.open(BytesIO(response.content)).convert("RGB")
    return image


class MyQwenVLModel(rlite.nn.HuggingFaceFsdp2TrainModule, Qwen2_5_VLForConditionalGeneration):
    def on_weights_materialized(self):
        # NOTE: This will place the processor on the worker process, which
        #       reduces the communication cost between the main process and
        #       the worker processes.
        model_name_or_path = self.materialization_source()
        self.processor = AutoProcessor.from_pretrained(model_name_or_path, use_fast=True)

        self.optimizer = torch.optim.AdamW(self.parameters(), lr=1e-5)

    def train_step(
        self,
        messages: list[list[dict]],
        image_urls: list[str],
        gradient_accumulation_steps: int = 1
    ):
        prompt_texts = [
            self.processor.apply_chat_template(
                messages,
                tokenize=False,
                add_generation_prompt=True,
            )
            for messages in messages
        ]
        images = [prepare_image(image_url) for image_url in image_urls]
        inputs = self.processor(
            text=prompt_texts,
            images=images,
            return_tensors="pt",
            padding=True,
        ).to("cuda")
        logits = self(**inputs).logits
        # Your loss function here, based on the logits
        loss = logits.mean() / gradient_accumulation_steps
        loss.backward()

    def optimizer_step(self):
        self.optimizer.step()
        self.optimizer.zero_grad()


if __name__ == "__main__":
    rlite.init()

    with init_empty_weights():
        config = AutoConfig.from_pretrained(
            "Qwen/Qwen2.5-VL-3B-Instruct",
            attn_implementation="flash_attention_2"
        )
        model = MyQwenVLModel(config)
        model.gradient_checkpointing_enable()

    engine = RliteTrainEngine(model, executor="fsdp2")
    engine.build(tensor_parallel_size=4)

    num_gpus = 8
    gradient_accumulation_steps = 4

    messages = [
        [
            {
                "role": "user",
                "content": [
                    {"type": "image", "image": "https://picsum.photos/id/1/200/300"},
                    {"type": "text", "text": "Describe the image in detail."}
                ]
            }
        ]
    ] * num_gpus
    image_urls = ["https://picsum.photos/id/1/200/300" for _ in range(num_gpus)]

    for _ in range(gradient_accumulation_steps):
        engine.train_step(
            NeedParallel(messages),
            NeedParallel(image_urls),
            gradient_accumulation_steps=gradient_accumulation_steps
        )
    engine.optimizer_step()



from __future__ import annotations

import abc

import ray
import torch

from rlite.utils.distributed import CUDAIPCHandle


class BaseWorker(abc.ABC):
    # ----- Interfaces ----- #

    @abc.abstractmethod
    def update_weight(self, key: str, value: torch.Tensor | CUDAIPCHandle | None):
        """Update a single weight of the worker.

        This is useful for sync the weights between colocated workers. The key
        can be "END" to indicate the end of the sync (useful for FSDP).
        """

    @abc.abstractmethod
    def sync_weights_to(self, worker: ray.ActorHandle[BaseWorker]):
        """Sync the weights to another worker."""

    @abc.abstractmethod
    def load_state_dict(self, state_dict: dict):
        """Load the state dict of the worker."""

    @abc.abstractmethod
    def state_dict(self) -> dict:
        """Get the state dict of the worker."""

    @abc.abstractmethod
    def cpu(self, *args, **kwargs):
        """Move the worker to the CPU."""

    @abc.abstractmethod
    def meta(self, *args, **kwargs):
        """Release all the weights (not including buffers)."""

    @abc.abstractmethod
    def cuda(self, *args, **kwargs):
        """Move the worker to the GPU."""

    @abc.abstractmethod
    def get_device(self) -> str:
        """Get the device of the worker."""

    # ----- Shared methods ----- #

    def release_memory_cache(self, cpu: bool = False, gpu: bool = True):
        if cpu:
            try:
                torch._C._host_emptyCache()
            except AttributeError:
                from loguru import logger

                logger.warning(
                    "torch._C._host_emptyCache() only available in Pytorch >=2.5"
                )
        if gpu:
            torch.cuda.empty_cache()

    def ready(self):
        """Check if the worker is ready."""
        return True



from __future__ import annotations

from typing import TYPE_CHECKING

from rlite.resman.resource_bundle import ResourceBundle

if TYPE_CHECKING:
    from rlite.resman.resource_manager import ResourceManager


class ResourceConsumer:
    """Interface for any class that consumes resources."""

    def __init__(self, resource_manager: ResourceManager | None = None):
        self._consumer_id = None
        self._resource_manager = None
        self._resource_bundles: list[ResourceBundle] = None

        if resource_manager is not None:
            resource_manager.register(self)

    def is_registered(self, manager: ResourceManager | None = None) -> bool:
        if not bool(self._consumer_id):  # not registered if "" or None
            return False
        if manager is not None and self._resource_manager is not manager:
            return False
        return True

    def is_resource_allocated(self) -> bool:
        return bool(self._resource_bundles)  # not allocated if [] or None

    def is_colocate_with(self, other: ResourceConsumer) -> bool:
        if self._resource_manager is not other._resource_manager:
            return False
        if len(self._resource_bundles) != len(other._resource_bundles):
            return False
        for bundle1, bundle2 in zip(self._resource_bundles, other._resource_bundles):
            if not bundle1.bundle_id_in_placement_group == bundle2.bundle_id_in_placement_group:
                return False
        return True



class AsyncRayContextManager:
    def __init__(self):
        self._is_in_context = False

    def __enter__(self):
        self._is_in_context = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._is_in_context = False

    def is_active(self):
        return self._is_in_context

    def __call__(self):
        return self



import abc

import torch

from rlite.nn.train_module import BaseHuggingFaceTrainModule, BaseTrainModule


class Fsdp2TrainModule(BaseTrainModule, abc.ABC):
    @property
    @abc.abstractmethod
    def non_split_modules(self) -> set[type[torch.nn.Module]]:
        """Get the modules that should not be split across FSDP instances."""
        raise NotImplementedError()

    def postprocess_output_named_parameters(
        self,
        name: str,
        param: torch.Tensor,
        full_state_dict: dict[str, torch.Tensor]
    ) -> tuple[str, torch.Tensor]:
        return name, param.full_tensor()


class HuggingFaceFsdp2TrainModule(BaseHuggingFaceTrainModule, Fsdp2TrainModule, abc.ABC):
    """Rlite FSDP Module for Hugging Face models."""



from enum import Enum


class ParallelType(Enum):
    DP = "dp"
    SEQ = "seq"


class NeedParallel:
    def __init__(self, data: any, /, *, type: str = "dp"):
        allowed_types = [
            "dp", "data parallel", "data_parallel", "data-parallel",
            "seq", "sequence", "sequence parallel", "sequence_parallel", "sequence-parallel",
            "ctx", "context", "context parallel", "context_parallel", "context-parallel",
        ]
        assert type in allowed_types, f"Invalid parallel type: {type}"
        if type in ["dp", "data parallel", "data_parallel", "data-parallel"]:
            self.parallel_type = ParallelType.DP
        else:
            self.parallel_type = ParallelType.SEQ

        self.data = data



from __future__ import annotations

import json
from pathlib import Path
from typing import Generator

import torch
from safetensors import safe_open
from safetensors.torch import save_file
from tqdm import tqdm


def load_safetensors_generator(
    safetensors_path,
    device="cpu",
    use_tqdm: bool = True,
    fetch_data: bool = True,
    dtype: torch.dtype = torch.float32,
):
    path = Path(safetensors_path)
    if not path.exists():
        raise FileNotFoundError(f"File not found: {safetensors_path}")

    if path.is_dir():
        paths = list(path.glob("*.safetensors"))
        if len(paths) == 0:
            raise FileNotFoundError(f"No safetensors files found in directory: {safetensors_path}")
    else:
        paths = [path]

    if use_tqdm:
        paths = tqdm(paths, desc="Loading safetensors")

    for safetensors_path in paths:
        with safe_open(safetensors_path, framework="pt", device=device) as f:
            for key in f.keys():
                if fetch_data:
                    yield key, f.get_tensor(key).to(dtype)
                else:
                    yield key, None


def save_safetensors(
    state_dict: dict | Generator,
    path: str,
    min_block_size_in_GB: int = 5,
    max_block_size_in_GB: int = 20,
    total_size: int | None = None,
    total_len: int | None = None,
    is_io_worker: bool = True,
):
    path = Path(path)
    path.mkdir(parents=True, exist_ok=True)

    if total_size is None and isinstance(state_dict, dict):
        total_size = 0
        for param in state_dict.values():
            total_size += param.numel() * param.element_size()

    if total_size is not None:
        size_per_block = total_size // 4 + 1
        size_per_block = min(size_per_block, max_block_size_in_GB * 1024**3)
        size_per_block = max(size_per_block, min_block_size_in_GB * 1024**3)
    else:
        from loguru import logger

        logger.warning(
            f"Total size is not provided, using min block size: {min_block_size_in_GB}GB."
        )
        size_per_block = min_block_size_in_GB * 1024**3

    if total_len is None and isinstance(state_dict, dict):
        total_len = len(state_dict)

    if isinstance(state_dict, dict):
        state_dict = state_dict.items()

    partial_state_dict = {}
    num_saved_files = 0
    partial_state_dict_size = 0
    safetensors_path = path / "model-1.safetensors"
    meta_info = {"metadata": {"total_size": total_size}, "weight_map": {}}

    for name, param in tqdm(
        state_dict,
        total=total_len,
        desc="Saving weights",
        disable=not is_io_worker,
    ):
        partial_state_dict[name] = param
        partial_state_dict_size += param.numel() * param.element_size()
        meta_info["weight_map"][name] = safetensors_path.name
        if partial_state_dict_size > size_per_block:
            if is_io_worker:
                save_file(partial_state_dict, safetensors_path, metadata={"format": "pt"})
            num_saved_files += 1
            safetensors_path = path / f"model-{num_saved_files + 1}.safetensors"

            partial_state_dict = {}
            partial_state_dict_size = 0

    if len(partial_state_dict) > 0 and is_io_worker:
        save_file(partial_state_dict, safetensors_path, metadata={"format": "pt"})

    if is_io_worker:
        with open(path / "model.safetensors.index.json", "w") as f:
            json.dump(meta_info, f)



from rlite.interface.train.train_executor import (
    TRAIN_EXECUTOR_REGISTRY,
    BaseTrainExecutor
)
from rlite.nn.fsdp import FsdpTrainModule
from rlite.resman import ResourceBundle
from rlite.train.fsdp.worker import FsdpWorker
from rlite.utils.misc import set_random_seed


@TRAIN_EXECUTOR_REGISTRY.register([
    "fsdp", "fsdp1", "FSDP", "FSDP1", "Fsdp", "Fsdp1", "FsdpTrainExecutor"
])
class FsdpTrainExecutor(BaseTrainExecutor):
    def __init__(
        self,
        *,
        dp_rank: int,
        module: FsdpTrainModule,
        bundles: list[ResourceBundle],
        tensor_parallel_size: int,
        data_parallel_size: int = 1,
        pipeline_parallel_size: int = 1,
        seed: int = 42,
        **fsdp_kwargs
    ):
        assert isinstance(module, FsdpTrainModule), type(module)
        self.dp_rank = dp_rank
        self.module = module
        self.bundles = bundles
        self.tensor_parallel_size = tensor_parallel_size
        self.data_parallel_size = data_parallel_size
        self.pipeline_parallel_size = pipeline_parallel_size
        self.seed = seed
        self.fsdp_kwargs = fsdp_kwargs

        self._workers: list[FsdpWorker] = []

    @property
    def workers(self):
        return self._workers

    def build_workers(self):
        self._workers = self._build_workers(
            module=self.module,
            dp_rank=self.dp_rank,
            bundles=self.bundles,
            data_parallel_size=self.data_parallel_size,
            tensor_parallel_size=self.tensor_parallel_size,
            pipeline_parallel_size=self.pipeline_parallel_size,
            seed=self.seed,
            **self.fsdp_kwargs
        )
        return [worker.ready.remote() for worker in self._workers]

    def _build_workers(
        self,
        module: FsdpTrainModule,
        dp_rank: int,
        bundles: list[ResourceBundle],
        data_parallel_size: int,
        tensor_parallel_size: int,
        pipeline_parallel_size: int,
        seed: int,
        replicate_master_addrs: list[str] | None = None,
        replicate_ports: list[int] | None = None,
        **fsdp_kwargs
    ):
        set_random_seed(seed)

        workers = []
        assert len(bundles) == tensor_parallel_size * pipeline_parallel_size, (
            f"The number of bundles ({len(bundles)=}) "
            "should be equal to the product of "
            f"tensor_parallel_size ({tensor_parallel_size=}) "
            f"and pipeline_parallel_size ({pipeline_parallel_size=})"
        )

        assert pipeline_parallel_size == 1, "Pipeline parallel is not supported yet!"
        assert tensor_parallel_size > 1, "FSDP requires tensor parallel size > 1"

        # Prepare the env variables
        shard_master_addr = bundles[0].resource_manager.get_ip_addr(bundles[0])
        shard_port = bundles[0].resource_manager.get_free_port(bundles[0])
        world_size = tensor_parallel_size * pipeline_parallel_size * data_parallel_size
        if data_parallel_size > 1:
            node_rank, num_nodes = dp_rank, data_parallel_size
        else:
            node_rank, num_nodes = 0, 1

        for rank in range(tensor_parallel_size * pipeline_parallel_size):
            bundle = bundles[rank]
            env_vars = {
                "RANK": rank,
                "WORLD_SIZE": world_size,
                "NODE_RANK": node_rank,
                "NUM_NODES": num_nodes,
                "SHARD_MASTER_ADDR": shard_master_addr,
                "SHARD_MASTER_PORT": shard_port,
            }
            if replicate_master_addrs is not None:
                env_vars["REPLICATE_MASTER_ADDR"] = replicate_master_addrs[rank]
                env_vars["REPLICATE_MASTER_PORT"] = replicate_ports[rank]
            env_vars = {k: str(v) for k, v in env_vars.items()}

            worker = bundle.remote(
                FsdpWorker,
                runtime_env={"env_vars": env_vars},
            ).remote(module=module, **fsdp_kwargs)
            workers.append(worker)

        return workers



import torch

from rlite.utils.need_parallel import NeedParallel, ParallelType


def _get_chunk_indices(data_size: int, parallel_size: int) -> list[int]:
    # Eg: len(arg)=7, dp_size=4 : [2, 2, 2, 1]
    chunk_sizes = [data_size // parallel_size] * parallel_size
    for i in range(data_size % parallel_size):
        chunk_sizes[i] += 1
    chunk_indeces = [0] * (parallel_size + 1)
    for i in range(1, len(chunk_indeces)):
        chunk_indeces[i] = chunk_sizes[i - 1] + chunk_indeces[i - 1]
    return chunk_indeces


def _handle_need_parallel(
    arg: NeedParallel,
    unwrap: bool = True,
    parallel_size: int = 1
) -> list[any]:
    if arg.parallel_type == ParallelType.DP:
        assert len(arg.data) >= parallel_size, "Data size smaller than parallel_size!"

        if isinstance(arg.data, torch.Tensor):
            data_chunks = [*torch.chunk(arg.data, parallel_size, dim=0)]
        else:
            chunk_indeces = _get_chunk_indices(len(arg.data), parallel_size)
            data_chunks = [
                arg.data[chunk_indeces[i]:chunk_indeces[i + 1]]
                for i in range(parallel_size)
            ]

        if not unwrap:
            data_chunks = [NeedParallel(data_chunk, type="dp") for data_chunk in data_chunks]

        return data_chunks

    elif arg.parallel_type == ParallelType.SEQ:
        assert len(arg.data[0]) >= parallel_size, "Sequence length smaller than parallel_size!"
        assert all(len(x) == len(arg.data[0]) for x in arg.data), (
            "All data chunks must have the same sequence length!"
        )

        if isinstance(arg.data, torch.Tensor):
            data_chunks = [*torch.chunk(arg.data, parallel_size, dim=1)]
        else:
            chunk_indeces = _get_chunk_indices(len(arg.data[0]), parallel_size)
            data_chunks = [
                [
                    arg.data[i][chunk_indeces[j]:chunk_indeces[j + 1]]
                    for i in range(len(arg.data))
                ]
                for j in range(parallel_size)
            ]

        if not unwrap:
            data_chunks = [NeedParallel(data_chunk, type="seq") for data_chunk in data_chunks]

        return data_chunks


def prepare_parallel_args(
    args: tuple,
    unwrap: bool = True,
    dp_size: int = 1
) -> list[list[any]]:
    if len(args) == 0:
        return [[]] * dp_size
    arg_list = []
    for arg in args:
        if isinstance(arg, NeedParallel):
            arg_list.append(_handle_need_parallel(arg, unwrap, dp_size))
        else:
            arg_list.append([arg] * dp_size)
    return [[arg_list[j][i] for j in range(len(arg_list))] for i in range(dp_size)]


def prepare_parallel_kwargs(
    kwargs: dict[str, any],
    unwrap: bool = True,
    dp_size: int = 1
) -> list[dict[str, any]]:
    if len(kwargs) == 0:
        return [{} for _ in range(dp_size)]
    new_kwargs = {}
    for key, value in kwargs.items():
        if isinstance(value, NeedParallel):
            new_kwargs[key] = _handle_need_parallel(value, unwrap, dp_size)
        else:
            new_kwargs[key] = [value] * dp_size
    return [{key: value[i] for key, value in new_kwargs.items()} for i in range(dp_size)]



from __future__ import annotations

import os
from functools import reduce
from pathlib import Path
from typing import Literal

import ray
import torch
import torch.distributed as dist
import torch.distributed.checkpoint as dcp
import torch.nn as nn
from torch.distributed.fsdp import FullStateDictConfig
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.distributed.fsdp import (
    MixedPrecision,
    ShardedStateDictConfig,
    ShardingStrategy,
    StateDictType
)

from rlite.interface import BaseWorker
from rlite.interface.train import BaseTrainWorker
from rlite.nn.fsdp import FsdpTrainModule
from rlite.train.utils.fsdp import (
    iter_fsdp1_state_dict,
    offload_model,
    offload_optimizer,
    reload_model,
    reload_optimizer
)
from rlite.utils.checkpoint import save_safetensors
from rlite.utils.distributed import (
    CUDAIPCHandle,
    destroy_process_group,
    init_process_group
)


class FsdpWorker(BaseTrainWorker):
    def __init__(self, module: FsdpTrainModule, *, device_id: int = 0):
        self._filter_warnings()
        self.model = self._init_model(module, device_id)

    def __getattr__(self, name):
        model = object.__getattribute__(self, "model")
        if name == "model":
            return model
        try:
            return super().__getattr__(name)
        except AttributeError:
            # NOTE: Here we hijack the getattr of the FSDP to make sure that
            #       the getattr of the model is called with `self` being the
            #       FSDP instance instead of the _fsdp_wrapped_module.
            try:
                return nn.Module.__getattr__(model, name)
            except AttributeError:
                if hasattr(model._fsdp_wrapped_module.__class__, name):
                    attr = getattr(model._fsdp_wrapped_module.__class__, name)
                    if callable(attr):
                        return attr.__get__(model)
                return getattr(model._fsdp_wrapped_module, name)

    def train(self):
        self.model.train()

    def eval(self):
        self.model.eval()

    def cpu(self):
        offload_model(self.model)
        if hasattr(self, "optimizer"):
            offload_optimizer(self.model.optimizer)
        self.release_memory_cache()

    def cuda(self, *args, **kwargs):
        reload_model(self.model)
        if hasattr(self.model, "optimizer"):
            reload_optimizer(self.model.optimizer)

    def meta(self):
        raise NotImplementedError("Currently not implemented, coming soon.")

    def sync_weights_to(self, target_worker: ray.ActorHandle[BaseWorker]):
        for key, value in iter_fsdp1_state_dict(self.model):
            ipc_handle = CUDAIPCHandle.from_tensor(value)
            future = target_worker.update_weight.remote(key, ipc_handle)
            ray.get(future)

        ray.get(target_worker.update_weight.remote("END"))

    def update_weight(
        self,
        key: str,
        weight: torch.Tensor | CUDAIPCHandle | None = None,
    ):
        if key == "END":
            with FSDP.state_dict_type(self.model, StateDictType.SHARDED_STATE_DICT):
                self.model.load_state_dict(self._weight_load_state_dict)
                delattr(self, "_weight_load_state_dict")
            return

        if not hasattr(self, "_weight_load_state_dict"):
            with FSDP.state_dict_type(self.model, StateDictType.SHARDED_STATE_DICT):
                self._weight_load_state_dict = self.model.state_dict()

        if isinstance(weight, CUDAIPCHandle):
            weight = weight.rebuild()
        if weight.device.type == "meta":
            weight = torch.empty_like(weight, device="cuda")
        if weight.device.type == "cpu":
            weight = weight.to("cuda", non_blocking=True)

        sharded_value = self._weight_load_state_dict[key]
        shard_dim = sharded_value._sharding_spec.dim
        v = torch.chunk(weight.cuda(), self.local_world_size, dim=shard_dim)[self.local_rank]
        sharded_value.local_shards()[0].tensor.copy_(v)

    def state_dict(self) -> dict:
        state_dict_config = FullStateDictConfig(offload_to_cpu=True, rank0_only=True)
        with FSDP.state_dict_type(self.model, StateDictType.FULL_STATE_DICT, state_dict_config):
            return self.model.state_dict()

    def load_state_dict(self, state_dict: dict):
        with FSDP.state_dict_type(self.model, StateDictType.FULL_STATE_DICT):
            self.model.load_state_dict(state_dict)

    @property
    def training(self) -> bool:
        return self.model.training

    def get_device(self) -> str:
        if not hasattr(self, "model"):
            raise RuntimeError(
                "FSDP worker is not initialized! You probably need to call "
                "`build_worker_groups` first."
            )
        return next(self.model.parameters()).device.type

    # See example here:
    # https://github.com/pytorch/pytorch/blob/main/torch/distributed/checkpoint/examples/fsdp_checkpoint_example.py  # noqa
    def save(
        self,
        checkpoint_path: str,
        type: Literal["distcp", "safetensors"] = "safetensors"
    ):
        if self.get_device() != "cuda":
            raise RuntimeError(
                f"Currently the model is on {self.get_device()}, you need to call "
                "`cuda()` first to save the checkpoint."
            )

        if type == "distcp":
            self._save_distcp(checkpoint_path)
        elif type == "safetensors":
            self._save_safetensors(checkpoint_path)

    def load(self, checkpoint_id: str, type: Literal["distcp", "safetensors", "auto"] = "auto"):
        if self.get_device() != "cuda":
            raise RuntimeError(
                f"Currently the model is on {self.get_device()}, you need to call "
                "`cuda()` first to load the checkpoint."
            )

        if type == "auto":
            # Determine the type of the checkpoint
            ckpt_path = Path(checkpoint_id)
            if any(p.is_file() for p in ckpt_path.glob("*.safetensors")):
                type = "safetensors"
            elif any(p.is_file() for p in ckpt_path.glob("*.distcp")):
                type = "distcp"

        valid_types = ["distcp", "safetensors", "auto"]
        if type not in valid_types:
            raise ValueError(f"Unknown checkpoint type: {type}, must be one of {valid_types}")

        if type == "safetensors":
            self._load_safetensors(checkpoint_id)
        elif type == "distcp":
            self._load_distcp(checkpoint_id)

    def _save_safetensors(self, checkpoint_id: str):
        with FSDP.state_dict_type(self.model, StateDictType.SHARDED_STATE_DICT):
            sharded_state_dict = self.model.state_dict()

        # Get the total number of parameters and the state dict size in Bytes
        total_len = len(sharded_state_dict)
        total_size, element_size = 0, None
        for value in sharded_state_dict.values():
            if element_size is None:
                element_size = torch.tensor([], dtype=value.dtype).element_size()
            numel = reduce(lambda x, y: x * y, value.shape)
            total_size += numel * element_size

        # Memory-efficiently save the state dict
        state_dict_iterator = iter_fsdp1_state_dict(sharded_state_dict=sharded_state_dict)
        save_safetensors(
            state_dict_iterator,
            checkpoint_id,
            total_size=total_size,
            total_len=total_len,
            is_io_worker=self.local_rank == 0,
        )

    def _save_distcp(self, checkpoint_id: str):
        with FSDP.state_dict_type(self.model, StateDictType.SHARDED_STATE_DICT):
            state_dict = {
                "model": self.model.state_dict(),
            }
            if hasattr(self, "optim"):
                state_dict["optim"] = FSDP.optim_state_dict(self.model, self.optimizer)
            dcp.save(
                state_dict,
                storage_writer=dcp.FileSystemWriter(checkpoint_id),
                process_group=self.shard_group
            )

    def _load_safetensors(self, checkpoint_id: str, model=None):
        from rlite.utils.checkpoint import load_safetensors_generator

        model = model or self.model
        current_model_device = next(model.parameters()).device.type

        if current_model_device != "cuda":
            raise RuntimeError("Currently the model is on CPU, so we cannot load the checkpoint.")

        with FSDP.state_dict_type(
            model,
            StateDictType.SHARDED_STATE_DICT,
            state_dict_config=ShardedStateDictConfig(offload_to_cpu=False)
        ):
            sharded_state_dict = model.state_dict()
            dtype = next(iter(sharded_state_dict.values())).dtype

        state_dict_iterator = load_safetensors_generator(
            checkpoint_id,
            device="cuda",
            use_tqdm=self.local_rank == 0,
            fetch_data=self.local_rank == 0,
            dtype=dtype
        )

        for key, value in state_dict_iterator:
            sharded_value = sharded_state_dict[key]
            value_shape = list(sharded_value.shape)
            value_shape[sharded_value._sharding_spec.dim] //= self.local_world_size
            recv_tensor = torch.empty(torch.Size(value_shape), device="cuda", dtype=dtype)
            if self.local_rank == 0:
                tensor_list = list(
                    torch.chunk(
                        value,
                        self.local_world_size,
                        dim=sharded_value._sharding_spec.dim
                    )
                )
            else:
                tensor_list = None
            dist.scatter(recv_tensor, tensor_list, src=0, group=self.shard_group)
            sharded_value.local_shards()[0].tensor.copy_(recv_tensor)

        with FSDP.state_dict_type(model, StateDictType.SHARDED_STATE_DICT):
            model.load_state_dict(sharded_state_dict)

    def _load_distcp(self, checkpoint_id: str):
        with FSDP.state_dict_type(self.model, StateDictType.SHARDED_STATE_DICT):
            state_dict = {
                "model": self.model.state_dict(),
            }
            dcp.load(
                state_dict,
                storage_reader=dcp.FileSystemReader(checkpoint_id),
                process_group=self.shard_group
            )
            self.model.load_state_dict(state_dict["model"])
            if hasattr(self, "optimizer"):
                from torch.distributed.checkpoint.optimizer import (
                    load_sharded_optimizer_state_dict
                )

                FSDP.optim_state_dict(self.model, self.optimizer, state_dict["optimizer"])
                optim_state = load_sharded_optimizer_state_dict(
                    model_state_dict=state_dict["model"],
                    optimizer_key="optim",
                    storage_reader=dcp.FileSystemReader(checkpoint_id),
                )
                flattened_osd = FSDP.optim_state_dict_to_load(
                    self.model, self.optimizer, optim_state["optim"]
                )
                self.optimizer.load_state_dict(flattened_osd)

    def _init_optimizer(
        self,
        optimizer_cls: type | str | None = None,
        learning_rate: float = 1e-5,
        optimizer_init_kwargs: dict | None = None,
    ):
        if isinstance(optimizer_cls, str):
            optimizer_cls = getattr(torch.optim, optimizer_cls)
        elif optimizer_cls is None:
            optimizer_cls = torch.optim.AdamW

        optimizer_init_kwargs = optimizer_init_kwargs or {}
        optimizer_init_kwargs["lr"] = learning_rate
        optimizer = optimizer_cls(self.model.parameters(), **optimizer_init_kwargs)
        return optimizer

    def _init_model(self, module: FsdpTrainModule, device_id: int = 0):
        # 1. Initialize process groups
        self.shard_group, self.replicate_group = self._init_process_groups()

        # 2. Determine the sharding strategy and process group
        if self.replicate_group is None:
            self.sharding_strategy = ShardingStrategy.FULL_SHARD
            process_group = self.shard_group
        else:
            self.sharding_strategy = ShardingStrategy.HYBRID_SHARD
            process_group = (self.shard_group, self.replicate_group)

        # 3. Materialize the model weights
        mat_src = module.materialization_source()
        if isinstance(mat_src, str):
            folder = Path(mat_src)
            wrap_policy = module.get_fsdp_wrap_policy()
            if len(list(folder.glob("*.safetensors"))) > 0:
                # TODO (zhanghan): support other checkpoint formats
                model = self._init_model_by_checkpoint_hf(
                    module, device_id, process_group, wrap_policy
                )
            else:
                model = self._init_model_by_checkpoint_distcp(
                    module, device_id, process_group, wrap_policy
                )
        elif isinstance(mat_src, dict):
            # For random initialization and train from scratch
            raise NotImplementedError("Currently not implemented, coming soon.")

        # 4. Set communication attributes for outside world
        model.shard_rank = model.module.shard_rank = self.local_rank
        model.shard_world_size = model.module.shard_world_size = self.local_world_size
        model.replicate_rank = model.module.replicate_rank = self.node_rank
        model.replicate_world_size = model.module.replicate_world_size = self.num_nodes

        # 4. Call hooks
        if hasattr(model, "on_weights_materialized"):
            model.on_weights_materialized()

        # 5. Explicitly set the model to training mode to avoid some bugs
        model.train()

        return model

    def _init_model_by_checkpoint_hf(self, module, device_id, process_group, wrap_policy):
        model = self._init_fsdp(
            module,
            device_id=device_id,
            process_group=process_group,
            sharding_strategy=self.sharding_strategy,
            auto_wrap_policy=wrap_policy,
        )

        self._load_safetensors(module.get_checkpoint_path(), model)

        return model

    def _init_model_by_checkpoint_distcp(self, module, device_id, process_group, wrap_policy):
        model = self._init_fsdp(
            module,
            device_id=device_id,
            process_group=process_group,
            sharding_strategy=self.sharding_strategy,
            auto_wrap_policy=wrap_policy,
        )
        with FSDP.state_dict_type(model, StateDictType.SHARDED_STATE_DICT):
            state_dict = {"model": model.state_dict()}
            dcp.load(
                state_dict,
                storage_reader=dcp.FileSystemReader(model.get_checkpoint_path()),
                process_group=self.shard_group,
            )
            model.load_state_dict(state_dict["model"])
            model = model.to(next(module.parameters()).dtype)
        return model

    def _init_fsdp(self, module, **kwargs):
        from rlite.train.utils.fsdp import param_init_preserve_buffers

        kwargs = dict(
            {
                "mixed_precision": MixedPrecision(
                    param_dtype=torch.bfloat16,
                    reduce_dtype=torch.float32,
                    buffer_dtype=torch.float32
                ),
                "param_init_fn": param_init_preserve_buffers,
                "forward_prefetch": True,
                "backward_prefetch": "BACKWARD_PRE",
                "use_orig_params": False,
                "device_id": "cuda",
            },
            **kwargs,
        )
        # For FSDP training model, float32 weights are required, otherwise the
        # performance can be problematic.
        return FSDP(module.float(), **kwargs)

    def _init_process_groups(self):
        # Collect distributed settings
        self.rank = int(os.getenv("RANK", 0))
        self.world_size = int(os.getenv("WORLD_SIZE", 1))
        self.node_rank = int(os.getenv("NODE_RANK", 0))
        self.num_nodes = int(os.getenv("NUM_NODES", 1))
        self.shard_master_addr = os.getenv("SHARD_MASTER_ADDR")
        self.shard_master_port = int(os.getenv("SHARD_MASTER_PORT"))

        self.local_world_size = self.world_size // self.num_nodes
        self.local_rank = self.rank % self.local_world_size

        # Initialize process groups
        shard_init_method = f"tcp://{self.shard_master_addr}:{self.shard_master_port}"
        conflict_free_shard_name = f"fsdp_shard_{self.shard_master_addr}_{self.shard_master_port}"
        dist.init_process_group(
            rank=self.local_rank,
            world_size=self.local_world_size,
            backend="nccl",
            group_name=conflict_free_shard_name,
            init_method=shard_init_method,
        )
        shard_group = dist.group.WORLD
        if self.num_nodes > 1:
            replicate_master_addr = os.getenv("REPLICATE_MASTER_ADDR")
            replicate_master_port = int(os.getenv("REPLICATE_MASTER_PORT"))
            replicate_init_method = f"tcp://{replicate_master_addr}:{replicate_master_port}"
            conflict_free_replicate_name = (
                f"fsdp_replicate_{replicate_master_addr}_{replicate_master_port}"
            )
            replicate_group = init_process_group(
                rank=self.node_rank,
                world_size=self.num_nodes,
                backend="nccl",
                group_name=conflict_free_replicate_name,
                init_method=replicate_init_method,
            )
            return shard_group, replicate_group
        return shard_group, None

    def _filter_warnings(self):
        import warnings

        warnings.filterwarnings("ignore", category=FutureWarning)
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message=".*_get_pg_default_device.*"
        )

    def __del__(self):
        if hasattr(self, "shard_group") and self.shard_group is not None:
            dist.destroy_process_group(self.shard_group)
        if hasattr(self, "replicate_group") and self.replicate_group is not None:
            destroy_process_group(self.replicate_group)

    # ----- Generator style state dict utilities ----- #

    # NOTE (zhanghan): these functions are very useful if we want to sync
    #                  weights more efficiently without CUDAIPC, while still
    #                  keeping the memory footprint low. These functions are
    #                  based on a fake generator implementation based on
    #                  message passing. The ray generator does not work, as it
    #                  does not pause the actor when the generator yields. It
    #                  will run the loop and save the yielded values in a
    #                  buffer. This can cause deaklock or messed up CUDA IPC.

    def _setup_state_dict_generator(self):
        with FSDP.state_dict_type(self.model, StateDictType.SHARDED_STATE_DICT):
            self._state_dict_iterator = iter(self.model.state_dict().items())

    def _generate_next_state(self):
        try:
            key, value = next(self.state_dict_iterator)
            gathered_value = torch.empty(
                value.shape,
                device=torch.device("cuda"),
                dtype=value.dtype,
            )
            value.gather(out=gathered_value if dist.get_rank() == 0 else None)
            if dist.get_rank() == 0:
                return key, gathered_value.cpu()
        except StopIteration:
            delattr(self, "_state_dict_iterator")
            self.clear_cache()
            return "END", None
        except AttributeError:
            return

    def _setup_state_dict_loader(self):
        """Load a full state dict."""
        with FSDP.state_dict_type(
            self.model,
            StateDictType.SHARDED_STATE_DICT,
            state_dict_config=ShardedStateDictConfig(offload_to_cpu=False)
        ):
            self._state_dict_loader = self.model.state_dict()

    def _load_next_state(
        self,
        key: str,
        value: torch.Tensor | None = None,
        broadcast_weight: bool = False,
    ):
        if key == "END":
            with FSDP.state_dict_type(self.model, StateDictType.SHARDED_STATE_DICT):
                self.model.load_state_dict(self._state_dict_loader)
                delattr(self, "_state_dict_loader")
                return

        # Reconstruct the weights
        if isinstance(value, CUDAIPCHandle):
            value = value.rebuild()
        if value.device.type == "meta":
            value = torch.empty_like(value, device="cuda")
        if value.device.type == "cpu":
            value = value.to("cuda", non_blocking=True)

        if broadcast_weight:
            dist.broadcast(value, src=0, group=self.shard_group)

        sharded_value = self._state_dict_loader[key]
        shard_dim = sharded_value._sharding_spec.dim
        v = torch.chunk(value, self.local_world_size, dim=shard_dim)[self.local_rank].cuda()
        sharded_value.local_shards()[0].tensor.copy_(v)



from transformers import AutoProcessor

import rlite
from rlite import RliteInferenceEngine


def prepare_image(image_url):
    from io import BytesIO

    import requests
    from PIL import Image

    response = requests.get(image_url)
    image = Image.open(BytesIO(response.content)).convert("RGB")
    return image


def prepare_prompts(processor, num_prompts: int = 8):
    image_urls = [f"https://picsum.photos/id/{i}/200/300" for i in range(num_prompts)]
    messages = [
        processor.apply_chat_template(
            [
                {
                    "role": "user",
                    "content": [
                        {"type": "image", "image": image_url},
                        {"type": "text", "text": "Describe the image in detail."}
                    ]
                }
            ],
            tokenize=False,
            add_generation_prompt=True,
        )
        for image_url in image_urls
    ]
    prompts = [
        {
            "prompt": messages[i],
            "multi_modal_data": {
                "image": prepare_image(image_urls[i])
            }
        }
        for i in range(num_prompts)
    ]
    return prompts


if __name__ == "__main__":
    rlite.init()

    model_name_or_path = "Qwen/Qwen2.5-VL-3B-Instruct"
    processor = AutoProcessor.from_pretrained(model_name_or_path, use_fast=True)
    prompts = prepare_prompts(processor, num_prompts=8)

    engine = RliteInferenceEngine(model_name_or_path, executor="vllm")
    engine.build(
        tensor_parallel_size=4,
        max_model_len=4096,
        max_num_seqs=1,
        mm_processor_kwargs={
            "min_pixels": 28 * 28,
            "max_pixels": 1280 * 28 * 28,
        },
        limit_mm_per_prompt={"image": 1, "video": 0, "audio": 0},
    )

    outputs = engine.generate(prompts, sampling_params={"temperature": 0.7, "seed": 42})
    for output in outputs:
        print("\033[92m", output.outputs[0].text, "\033[0m")



import ray
from vllm import SamplingParams

from rlite.inference.vllm.worker import VllmWorker
from rlite.interface.inference.inference_executor import (
    INFERENCE_EXECUTOR_REGISTRY,
    BaseInferenceExecutor
)
from rlite.resman import ResourceBundle
from rlite.utils.misc import set_random_seed


@INFERENCE_EXECUTOR_REGISTRY.register(names=["vllm", "vLLM", "VLLM", "VllmInferenceExecutor"])
class VllmInferenceExecutor(BaseInferenceExecutor):
    def __init__(
        self,
        *,
        model_path: str,
        bundles: list[ResourceBundle],
        tensor_parallel_size: int = 1,
        pipeline_parallel_size: int = 1,
        seed: int = 42,
        # env related arguments
        use_v1_engine: bool = True,
        enable_vllm_logs: bool = False,
        **vllm_kwargs
    ):
        self.model_path = model_path
        self.bundles = bundles
        self.tensor_parallel_size = tensor_parallel_size
        self.pipeline_parallel_size = pipeline_parallel_size
        self.seed = seed
        self.use_v1_engine = use_v1_engine
        self.enable_vllm_logs = enable_vllm_logs
        self.vllm_kwargs = vllm_kwargs

        self._workers: list[VllmWorker] = []

    def build_workers(self):
        self._workers: list[VllmWorker] = self._build_workers(
            model_path=self.model_path,
            bundles=self.bundles,
            tensor_parallel_size=self.tensor_parallel_size,
            pipeline_parallel_size=self.pipeline_parallel_size,
            seed=self.seed,
            use_sleep_mode=True,
            use_v1_engine=self.use_v1_engine,
            enable_vllm_logs=self.enable_vllm_logs,
            **self.vllm_kwargs
        )
        return [worker.ready.remote() for worker in self._workers]

    @property
    def workers(self) -> list[VllmWorker]:
        return self._workers

    def generate(
        self,
        prompts: list[str],
        sampling_params: dict | SamplingParams | None = None,
        use_tqdm: bool = True,
        tqdm_desc: str = "Processed prompts",
        return_none: bool = False,
    ) -> list[ray.ObjectRef]:
        return [
            worker.generate.remote(prompts, sampling_params, use_tqdm, tqdm_desc, return_none)
            for worker in self.workers
        ]

    def p2p_weight_sync(self, another: BaseInferenceExecutor):
        raise NotImplementedError(
            "Currently VllmInferenceExecutor does not support p2p weight sync!"
        )

    def state_dict(self):
        raise NotImplementedError("Currently VllmInferenceExecutor does not support state_dict()")

    def load_state_dict(self, state_dict: dict):
        futures = [
            worker.load_state_dict.remote(state_dict)
            for worker in self.workers
        ]
        return futures

    def _build_workers(
        self,
        model_path: str,
        bundles: list[ResourceBundle],
        tensor_parallel_size: int,
        pipeline_parallel_size: int,
        seed: int,
        use_v1_engine: bool,
        enable_vllm_logs: bool,
        **vllm_kwargs
    ) -> list[VllmWorker]:
        set_random_seed(seed)
        workers = []
        assert len(bundles) == tensor_parallel_size * pipeline_parallel_size, (
            f"The number of bundles ({len(bundles)=}) "
            "should be equal to the product of "
            f"tensor_parallel_size ({tensor_parallel_size=}) "
            f"and pipeline_parallel_size ({pipeline_parallel_size=})"
        )

        # TODO (zhanghan): handle pipeline parallel later
        assert pipeline_parallel_size == 1, "Pipeline parallel is not supported yet!"

        master_addr = bundles[0].resource_manager.get_ip_addr(bundles[0])
        # NOTE: there are chances that the port is already used by other
        #       processes, as we don't lock the port.
        master_port = bundles[0].resource_manager.get_free_port(bundles[0])

        for rank in range(tensor_parallel_size * pipeline_parallel_size):
            bundle = bundles[rank]

            env_vars = {
                "RANK": str(rank),
                "WORLD_SIZE": len(bundles),
                "LOCAL_RANK": "0",
                "MASTER_ADDR": master_addr,
                "MASTER_PORT": master_port,
                "VLLM_USE_V1": int(use_v1_engine),
                "VLLM_CONFIGURE_LOGGING": int(enable_vllm_logs),
            }
            env_vars = {k: str(v) for k, v in env_vars.items()}

            worker = bundle.remote(
                VllmWorker,
                runtime_env={"env_vars": env_vars}
            ).remote(
                model_path=model_path,
                seed=seed,
                tensor_parallel_size=tensor_parallel_size,
                pipeline_parallel_size=pipeline_parallel_size,
                **vllm_kwargs
            )
            workers.append(worker)

        return workers



from rlite.train.engine import RliteTrainEngine
from rlite.train.fsdp import FsdpTrainExecutor
from rlite.train.fsdp2 import Fsdp2TrainExecutor

__all__ = [
    "RliteTrainEngine",
    "FsdpTrainExecutor",
    "Fsdp2TrainExecutor"
]



import torch
import torch.nn.functional as F
from accelerate import init_empty_weights
from transformers import Qwen2ForCausalLM

import rlite
import rlite.nn
from rlite import NeedParallel, RliteInferenceEngine, RliteTrainEngine


class MyQwenModel(rlite.nn.HuggingFaceFsdp2TrainModule, Qwen2ForCausalLM):
    def on_weights_materialized(self):
        self.optimizer = torch.optim.AdamW(self.parameters(), lr=1e-5)

    def train_step(self, input_ids: list[list[int]], grad_acc_steps: int = 1):
        input_ids = torch.tensor(input_ids, dtype=torch.long, device="cuda")

        # Language loss
        logits = self(input_ids).logits[:, :-1].contiguous()
        labels = input_ids[:, 1:].contiguous()
        loss = F.cross_entropy(logits.view(-1, logits.size(-1)), labels.view(-1)).mean()

        # Backward and gradient accumulation
        loss = loss / grad_acc_steps
        loss.backward()

        return loss.item()

    def optimizer_step(self):
        self.optimizer.step()
        self.optimizer.zero_grad()


if __name__ == "__main__":
    rlite.init()
    vllm_engine = RliteInferenceEngine("Qwen/Qwen2.5-7B-Instruct", executor="vllm")
    vllm_engine.build(tensor_parallel_size=4)

    prompts = ["你好，世界！", "Hello, world!"] * 8
    rollouts = vllm_engine.generate(prompts)

    vllm_engine.meta()  # Release everything from GPU

    with init_empty_weights():
        module = MyQwenModel.from_pretrained("Qwen/Qwen2.5-7B-Instruct")
    fsdp2_engine = RliteTrainEngine(module, executor="fsdp2")
    fsdp2_engine.build(tensor_parallel_size=4, colocate_with=vllm_engine)

    input_ids = [[x.outputs[0].token_ids for x in rollouts] for _ in range(4)]
    grad_acc_steps = 4

    for step in range(grad_acc_steps):
        fsdp2_engine.train_step(
            NeedParallel(input_ids[step]),  # This will be split among the workers (DP)
            grad_acc_steps  # This will be copied to all workers
        )
    fsdp2_engine.optimizer_step()

    vllm_engine.cuda("weights")  # Only the weights are loaded
    fsdp2_engine.p2p_weight_sync(vllm_engine)  # GPU-to-GPU weight sync via CUDAIPC
    fsdp2_engine.cpu()
    vllm_engine.cuda("kv_cache")  # The KV cache of vLLM is back to GPU
